#include <stdlib.h>
#include <string.h>
#include <poll.h>
#include "vr.H"
#include "tools.H"
#include "lob.H"
#include "forwarder.H"
#include "api.H"
#include "ping.H"
#include "socks.H"
#include "vrevent.H"
LOBSTART(Ping);

Ping::Ping(LOBenv *e) : LOB(e), VRschedObj(), ForwarderIF() 
{
  rtt_min=100000;
  rtt_max=0;
  rtt_avg=0;
  d_addr=0;
  verbose=0;
  ping_rep=3;			// ping three times
  ping_maxwait=10;		// stop after 10 seconds
  cnt=0;
  rcnt=0;
  for(int i=0;i<env->argc;++i){
    char *t=env->argv[i];
    if(t[0]!='-'){
      if(d_addr){
	d_addr=0;
	show_help();
	fhprintf(*env->cfh,"Error: wrong syntax\n");
	return;
      }
      d_addr=char2n(env->argv[i]);
      if(d_addr==0xffffffff){
	d_addr=0;
	show_help();
	fhprintf(*env->cfh,"Error: malformed destination address\n");
	return;
      }
      continue;
    }

    if(t[1]=='v'){
      verbose++;
      continue;
    }
    if(t[1]=='n'){
      if(i+1>=env->argc){
	show_help();
	fhprintf(*env->cfh,"Error: -n lacks parameter\n");
	d_addr=0;
	return;
      }
      ++i;
      ping_rep=atoi(env->argv[i]);
      if(!ping_rep){
	show_help();
	fhprintf(*env->cfh,"Error: malformed argument\n");
	return;
      }
      if(ping_rep>60){
	ping_rep=60;
	fhprintf(*env->cfh,"Warning: number of repetitions stripped to 60\n");
	continue;
      }
    }
  }

  if(!d_addr){
    show_help();
    return;
  }
 

  //  env->set_mode(LOB_MODE_RUN);	

  if(d_addr){
    ping_maxwait=ping_rep+3;

    ping_id=42; //(unsigned short)(gettimeasdouble()*1000);
    ping_sn=0;
    env->set_mode(LOB_MODE_STAY);
    
    pinghash=(double*)malloc(sizeof(double)*256);
    memset(pinghash,0,sizeof(double)*256);
    
    event=new VRevent(this);
    event->type=TIMER;
    event->timer_time=1.0;
    env->EventNotify(event);

    // communicatio with raw sockets

    // incoming packets

    sin = new RawSock(VRSOCK_READ);
    // we want ICMP
    sin->setopt(VRSOCK_PROTO,1);
    // with headers please
    sin->setopt(VRSOCK_HDRINC,1);
    // do not use the pipe, call out forward function
    sin->setopt(VRSOCK_DIRECT,this);

    // outgoing packets

    sout = new RawSock(VRSOCK_WRITE);
    // we generate ICMP packets
    sout->setopt(VRSOCK_PROTO,1);
    // set the destination address
    sout->setopt(VRSOCK_DEST,d_addr);
    // the host shall generate the header
    sout->setopt(VRSOCK_HDRINC,0); 
  }

  return;
}

void Ping::show_help()
{
  fhprintf(*env->cfh,"ping for Virtual Routers (c) baumgart@iam.unibe.ch\n");
  fhprintf(*env->cfh,"Syntax: ping [-n repetitions] [-v] destination\n");
  return;
}


/*
 * This is called by a timer event every second
 */

int Ping::interrupt(VRevent *e)
{
  if(cnt<ping_rep){
    if(verbose)fhprintf(*env->cfh,"interrupt: pinging to %s\n",n2char(d_addr));
    send_a_ping();
  }
  
  if(cnt>=ping_maxwait || rcnt>=ping_rep){
    // we are finished

    fhprintf(*env->cfh,"%d packets transmitted, %d packets received\n",
	     ping_rep, rcnt);
    if(rcnt){
      rtt_avg/=rcnt;
      fhprintf(*env->cfh,"round-trip min/avg/max = %4.1f/%4.1f/%4.1f ms\n", 
	       rtt_min, rtt_avg, rtt_max);
    }

    // unregistering Event
    env->EventRelease(event->event_id);
    delete event;

    // closing sockets
    
    delete sout;
    delete sin;
    env->finish();
    
    return 0;
  }
  ++cnt;
  return 0;
}

/*
 * This is called if a packet arrives
 */

int Ping::forward(Sock *s, Packet *p)
{
  struct icmp_header *pic=(struct icmp_header*)PKT_DATA(p);
  

  // if packet is not ICMP echo reply, forget it
  if(pic->type!=0)return 0;

  // fhprintf(*env->cfh,"%s\n",pkt2char(p));
  // fhprintf(*env->cfh,"%d %d\n",(int)pic->type, (int)pic->code);

  double rtt=1000*(gettimeasdouble()-pinghash[pic->sn%256]);
  if(rtt<rtt_min)rtt_min=rtt;
  if(rtt>rtt_max)rtt_max=rtt;
  rtt_avg+=rtt;
  rcnt++;
  fhprintf(*env->cfh,"%d bytes from %s: icmp_seq=%d icmp_id=%d ttl=%d time=%4.1f ms\n",
	   PKT_DATALEN(p), n2char(PKT_SOURCE(p)), pic->sn, pic->id, PKT_TTL(p), rtt);
  return 0;
} 


Ping::~Ping() 
{
  return;
}


int Ping::send_a_ping()
{
  char buffer[128];
  struct icmp_header *pic=(struct icmp_header*)buffer;
  memset(buffer,0,128);
  pic->type=8;
  pic->code=0;
  pic->id=ping_id; 
  pic->sn=ping_sn++;
  pic->cksum=cksum(buffer,64);
  pinghash[pic->sn%256]=gettimeasdouble();
  sout->write(buffer,64);
  return 0;
}




