#include <forwarder.H>
#include <tools.H>
#include <Python.h>
#include <lob.H>
#include "pybar.H"
#include "pad.H"
#include "v_router.H"

static PyObject *pad_version(PyObject*, PyObject *);
static PyObject *pad_Router(PyObject*, PyObject *);
static PyObject *pad_Packet(PyObject*, PyObject *);
static PyObject *pad_getpacket(PyObject*, PyObject*);

static void Router_dealloc(PyObject *);
static PyObject* Router_getattr(PyObject*, char *);
static PyObject* Router_init(PyObject *, PyObject*);

static void Packet_dealloc(PyObject *);
static PyObject* Packet_getattr(PyObject*, char *);


/*
 * Methods of the pad module
 */


static PyMethodDef pad_methods[] = {
  {"version",		pad_version, 1},
  {"Router",            pad_Router,  1},
  {"Packet",            pad_Packet,  1}, 
  {"getpacket",         pad_getpacket, 1},
  {NULL,		NULL}		// sentinel 
};

/*
 * Methods of the Router Object
 */


static PyMethodDef Router_methods[] = {
  { "init", Router_init, 1 },
  {NULL, NULL}
}; 


/*
 * Router Object and Router Type
 */

typedef struct {
  PyObject_HEAD
  Router *R;                    
} RouterObject; 


static PyTypeObject RouterType = {
  PyObject_HEAD_INIT(NULL)
  0,
  "Router",
  sizeof(RouterObject),
  0,
  Router_dealloc,               /*tp_dealloc*/
  0,                            /*tp_print*/
  Router_getattr,               /*tp_getattr*/
  0,                            /*tp_setattr*/
  0,                            /*tp_compare*/
  0,                            /*tp_repr*/
  0,                            /*tp_as_number*/
  0,                            /*tp_as_sequence*/
  0,                            /*tp_as_mapping*/
  0,                            /*tp_hash */
}; 

/*
 * packet object and type
 */

typedef struct {
  PyObject_HEAD
  Packet *p;
} PacketObject; 

static PyTypeObject PacketType = {
  PyObject_HEAD_INIT(NULL)
  0,
  "Packet",
  sizeof(PacketObject),
  0,
  Packet_dealloc,               /*tp_dealloc*/
  0,                            /*tp_print*/
  Packet_getattr,               /*tp_getattr*/
  0,                            /*tp_setattr*/
  0,                            /*tp_compare*/
  0,                            /*tp_repr*/
  0,                            /*tp_as_number*/
  0,                            /*tp_as_sequence*/
  0,                            /*tp_as_mapping*/
  0,                            /*tp_hash */
}; 


/*
 * pad.version()
 */

static PyObject *pad_version(PyObject *self , PyObject *args)
{
  if (!PyArg_ParseTuple(args, ""))return NULL;
  return Py_BuildValue("(s,d)",PAD_PLATFORM,PAD_VERSION);
}

static PyObject *pad_getpacket(PyObject *self , PyObject *args)
{
  Pybar *pyb=NULL;
  double ti_int;
  double ti_end=0.0;

  if (!PyArg_ParseTuple(args, "d", &ti_int))return NULL;

  ti_end=ti_int+gettimeasdouble();

  PyObject *m = PyImport_ImportModule("pad"); 
  if (m != NULL) { 
    PyObject *module_dict = PyModule_GetDict(m); 
    PyObject *pobject = PyDict_GetItemString(module_dict, "pybar_lob_ptr");
    if (PyCObject_Check(pobject)) { 
      pyb=(Pybar*)PyCObject_AsVoidPtr(pobject);
    } 
  } 
 
  if(!pyb)return NULL;		// should NEVER happen

  // okay, here we go, we've got all references

  while(1){
    //
    // if packet in queue process it and return
    //
    if(pyb->packets_in_iqueue){	
      Packet *pkt=pyb->iqueue[pyb->iq_idx_out];
      if(!pkt)return NULL;		// should NEVERT happen
      pyb->iqueue[pyb->iq_idx_out]=NULL;
      pyb->iq_idx_out++; 
      if(pyb->iq_idx_out>=pyb->iqueue_len)pyb->iq_idx_out=0;
      
      PacketObject *po=PyObject_New(PacketObject, &PacketType);
      if(!po){			// should NEVER happen
	free(pkt);
	return NULL;
      }
      po->p=pkt;
      return (PyObject*)po;
    }
    if(gettimeasdouble()>=ti_end){
      break;			// time exceeded
    }
    usleep(1);
  }
  
  //
  // return None Object
  //

  PyObject *po=Py_None;
  Py_INCREF(po);

  return (PyObject*)po; 
}


/*
 *  k=pad.Router()
 */

static PyObject *pad_Router(PyObject *self, PyObject *args)
{
  RouterObject *r=NULL;
  Pybar *pyb=NULL;

  if (!PyArg_ParseTuple(args,":Router"))
    return NULL;     
  
  PyObject *m = PyImport_ImportModule("pad"); 
  if (m != NULL) { 
    PyObject *module_dict = PyModule_GetDict(m); 
    PyObject *pobject = PyDict_GetItemString(module_dict, "pybar_lob_ptr");
    if (PyCObject_Check(pobject)) { 
      pyb=(Pybar*)PyCObject_AsVoidPtr(pobject);
    } 
  } 
 
  if(!pyb)return NULL;
 
  r = PyObject_New(RouterObject, &RouterType);     
  r->R=new Router(pyb);

  return (PyObject*)r;
}

static PyObject *pad_Packet(PyObject *self, PyObject *args)
{
  PacketObject *p=NULL;

  if (!PyArg_ParseTuple(args,":Packet"))
    return NULL;     
  
  p = PyObject_New(PacketObject, &PacketType);     
  p->p=NULL;

  return (PyObject*)p;
}



/*
 * removal of python object
 */

static void Router_dealloc(PyObject* self)
{
  delete ((RouterObject*)self)->R;
  PyObject_Del(self);
} 

/*
 * removeal of packet object
 */

static void Packet_dealloc(PyObject* self)
{
  if(((PacketObject*)self)->p)free(((PacketObject*)self)->p);
  PyObject_Del(self);
} 

static PyObject* Packet_getattr(PyObject* self, char *name)
{
  Packet *p=((PacketObject*)self)->p;
  
  if(!strcmp(name,"isempty")){
    if(!p)return PyInt_FromLong(1);
    return PyInt_FromLong(0);
  }
  
  if(!p)return NULL;

  if(!strcmp(name,"source"))
    return PyString_FromString(n2char(PKT_SOURCE(p)));
  
  if(!strcmp(name,"proto"))
    return PyInt_FromLong(PKT_PROTO(p));
  
  if(!strcmp(name,"dest"))
    return PyString_FromString(n2char(PKT_DEST(p)));

  if(!strcmp(name,"len"))
    return PyInt_FromLong(PKT_LENGTH(p));
  
  if(!strcmp(name,"payload"))
    return PyString_FromStringAndSize(PKT_DATA(p), PKT_DATALEN(p));

  return NULL;
} 


/*
 * getattr for the router object
 */

static PyObject* Router_getattr(PyObject* self, char *name)
{
  /*
    if(!strcmp(name,"id")){
    return PyInt_FromLong(((RouterObject*)self)->R->get_id());
    }
  */

  return Py_FindMethod(Router_methods, self, name);
} 


/*
 * The init method
 */

static PyObject* Router_init(PyObject* self, PyObject* args)
{
  if (!PyArg_ParseTuple(args,""))
    return NULL;     
  
  int r=((RouterObject*)self)->R->init();
  return PyInt_FromLong(r);
} 



void init_pad(Pybar *pyb)
{
  PyImport_AddModule("pad");
  PyObject *m=Py_InitModule("pad", pad_methods);
  
  /*
    okay nun stopfen wir notwenige Referenzen in den pad namespace 
  */

  PyObject *p=PyCObject_FromVoidPtr((void *)pyb, NULL);
  if(p != NULL) {
    /* Create a name for this object in the module's namespace */
    PyObject *d = PyModule_GetDict(m);         
    PyDict_SetItemString(d, "pybar_lob_ptr", p);
    Py_DECREF(p);
  }

  return;
}

