#include <stdlib.h>
#include <string.h>
#include "vr.H"
#include "packet.H"
#include "vrevent.H"
#include "tools.H"
#include "lob.H"
#include "forwarder.H"
#include "traceroute.H"
#include "socks.H"

LOBSTART(TraceRoute);

TraceRoute::TraceRoute(LOBenv *e) : LOB(e)
{
  d_addr=0;
  verbose=0;

  if(env->argc!=1){
    show_help();
    return;
  }

  d_addr=char2n(env->argv[0]);
  if(d_addr==0xffffffff){
    d_addr=0;
    show_help();
    fhprintf(*env->cfh,"Error: malformed destination address\n");
    return;
  }

  /*
   * Traceroute's interrupt is called every second
   */

  event=new VRevent(this);
  event->type=TIMER;
  event->timer_time=1.0;
  env->EventNotify(event);
  
  env->set_mode(LOB_MODE_STAY);
  
  /*
   * socket for incoming raw IP icmp packets
   */

  sin = new RawSock(VRSOCK_READ);
  sin->setopt(VRSOCK_PROTO,1);
  sin->setopt(VRSOCK_HDRINC,1);	// we want the complete IP packet
  // do not use the pipe, call out forward function
  sin->setopt(VRSOCK_DIRECT,this);

  /*
   * socket for outgoing UDP test packets
   */
  
  sout = new UDPSock(VRSOCK_WRITE);
  sout->setopt(VRSOCK_DEST,d_addr);
  sout->setopt(VRSOCK_REMOTEPORT,33434);

  hops=1;

  for(int i=0;i<128;++i)testpattern[i]=0x42; //rand()%256;

  return;
}


/*
 * This is called if a packet arrives
 */

int TraceRoute::forward(Sock *s, Packet *p)
{
  Packet *oh;
  struct icmp_header *pic=(struct icmp_header*)PKT_DATA(p);
  oh=(Packet*)&pic->data;

  fhprintf(*env->cfh,"TRACEROUTE: %s\n", pkt2char(p));
  fhprintf(*env->cfh,"TRACEROUTE: (%s)\n", pkt2char(oh));

  //  if(PKT_DEST(oh)!=d_addr || PKT_PROTO(oh)!=17)return 0;

  //  fhprintf(*env->cfh,"%s PKT_TTL(oh)

  /*
    double rtt=1000*(gettimeasdouble()-pinghash[pic->sn%256]);
    if(rtt<rtt_min)rtt_min=rtt;
    if(rtt>rtt_max)rtt_max=rtt;
    rtt_avg+=rtt;
    rcnt++;
    
    fhprintf(*env->cfh,"%d bytes from %s: icmp_seq=%d icmp_id=%d ttl=%d time=%4.1f ms\n",
    PKT_DATALEN(p), n2char(PKT_SOURCE(p)), pic->sn, pic->id, PKT_TTL(p), rtt);
  */
  return 0;
} 



void TraceRoute::show_help()
{
  fhprintf(*env->cfh,"traceroute for Virtual Routers (c) baumgart@iam.unibe.ch\n");
  fhprintf(*env->cfh,"Syntax: traceroute destination\n");
  return;
}


/*
 * This is called by a timer event every second
 */

int TraceRoute::interrupt(VRevent *e)
{
  if(hops<5){
    sout->setopt(VRSOCK_TTL,hops);
    sout->write(testpattern,128);
    fhprintf(*env->cfh,"sending traceroute test packet\n");
  }

  if(hops>10){
    // unregistering Event
    env->EventRelease(event->event_id);
    delete event;
    
    // closing sockets
    
    delete sout;
    delete sin;
    env->finish();
  }
  ++hops;
  return 0;
}
  
TraceRoute::~TraceRoute() 
{
  return;
}


