#ifndef _api_interface_h_
#define _api_interface_h_

//> QS_STATISTICS_VRRB
struct QS_STATISTICS_VRRB : public VRRB {
  unsigned short component_id;
  unsigned short component_type;
};
//<

//> QS_DROPTAIL_STATISTICS_VRRB
struct QS_DROPTAIL_STATISTICS_VRRB : public QS_STATISTICS_VRRB {
  unsigned short queuelen;
  unsigned long dropped_packets;
  unsigned long deque_requests;
  unsigned long deque_hits;
  unsigned long deque_fails;
};
//<

//> QS_TRIO_STATISTICS_VRRB
struct QS_TRIO_STATISTICS_VRRB : public QS_STATISTICS_VRRB {
  unsigned short mode;		       //- allows mode the TRIO queue is currently in
  unsigned short queuelen[3];	       //- the queue lengths for the three precedences
  unsigned short pkts;		       //- packets in queue
  unsigned long dropped_packets[3];    //- the number of drops for the three queuelengths
  unsigned long deque_requests;
  unsigned long deque_hits;
  unsigned long deque_fails;
};
//<

//> QS_CLASSIFIER_STATISTICS_VRRB
struct QS_CLASSIFIER_STATISTICS_VRRB : public QS_STATISTICS_VRRB {
  unsigned long deques;
  unsigned long enques;
};
//<

//> QS_DSMARKER_STATISTICS_VRRB
struct QS_DSMARKER_STATISTICS_VRRB : public QS_STATISTICS_VRRB {
  unsigned long enques;		//- \# packets all over
  unsigned long newmarks;	//- \# new mark set
  unsigned long oldmarks;	//- \# same mark as before
};
//<

//> DS_MARKER_RECORD
struct DS_MARKER_RECORD
{
  unsigned long source_ip;	//- the source ip address
  unsigned long source_nm;	//- the source netmask
  unsigned long dest_ip;	//- the destination address
  unsigned long dest_nm;	//- the destination netmask
  unsigned short tos;		//- the tos value
  unsigned short proto;		//- the protocol 
  unsigned short service;	//- the service
  unsigned long param[4];	//- additional parameters (e.g. for AF)
};
//<

//> QS_SCHEDULER_STATISTICS_VRRB
struct QS_SCHEDULER_STATISTICS_VRRB : public QS_STATISTICS_VRRB {
  unsigned short type;		//- type of the outgoing comp.
  unsigned short id;		//- id of the outgoing comp.
  unsigned short mode;		//- scheduling mode
};
//<

//> GEN_SCHEDULER_RECORD
struct GEN_SCHEDULER_RECORD
{
  unsigned short type;		//- type of the component
  unsigned short id;		//- id of the component
  unsigned long deque_hits;	//- successdul deques
  unsigned short misc;		//- depends on sched type
};
//<

//> GEN_CLASSIFIER_RECORD
struct GEN_CLASSIFIER_RECORD
{
  unsigned short type;		//- type of the outgoing
  unsigned short id;		//- id of the outgoing
  unsigned long source_ip;	//- the source ip address
  unsigned long source_nm;	//- the source netmask
  unsigned long dest_ip;	//- the destination address
  unsigned long dest_nm;	//- the destination netmask
  unsigned short tos;		//- the tos value
  unsigned short proto;		//- the protocol 
};
//<


//> QS_ROOT_STATISTICS_VRRB
struct QS_ROOT_STATISTICS_VRRB : public QS_STATISTICS_VRRB {
  unsigned long enques;
  unsigned long deques;
  //  unsigned long packets_in_qs;
  unsigned long global_stored;
  unsigned long global_dropped;
};
//<

//> QS_TBF_STATISTICS_VRRB
struct QS_TBF_STATISTICS_VRRB : public QS_STATISTICS_VRRB {
  unsigned long bandwidth;
  unsigned long bucketsize;
  unsigned long enque_hits;
  unsigned long enque_fails;
  unsigned long enque_empty;
  unsigned long deque_hits;
  unsigned long deque_fails;
  unsigned long deque_empty;
  unsigned long bucket_overflows;
};
//<


//> IF_ADDINTERFACE_CB
//c Control Block structure used for the creation of a new interface
struct IF_ADDINTERFACE_CB : public VRCB 
{
  char name[10];		//- interface name. The String 
				//- has to be followed by a 0-byte
  unsigned long ip;		//- ip address in network byte order        
  unsigned long nm;		//- netmask in network byte order
  unsigned long bc;		//- broadcast address in network byte order
};
//<

//> IF_IDNAMEQUERY_CB
struct IF_IDNAMEQUERY_CB : public VRCB
{
  char name[10];		//- interface name
};
//<

//> IF_DELETEINTERFACE_CB
//c Control Block to be send to the API for for 
//c the removal of an interface
struct  IF_DELETEINTERFACE_CB : public VRCB
{
  unsigned short ifno;		//- interface number
};
//<

//> IF_DELETEINTERFACE_RB
//c Result Block Block replied by the API
struct  IF_DELETEINTERFACE_RB : public VRRB
{
  unsigned short ifno;		//- interface number
};
//<

//> IF_VRCB
//c Basic control block for interface specific commands
struct IF_VRCB : public VRCB 
{
  unsigned short ifno;		//- interface number
  unsigned short cmd;		//- command to be processed by the interface
};
//<

//> IF_Q_VRCB
//c basic control block used for the configuration of an interface's 
//c queuing system. The rest of the CB is command specific.
struct IF_Q_VRCB : public IF_VRCB 
{
  unsigned short qcmd;		//- queueing command id
};
//<

//> IF_QS_VRCB
struct IF_QS_VRCB : public IF_Q_VRCB 
{
  unsigned short component_id;	//- id of the component to be accessed
  unsigned short component_command; //- command code specific for the component
};
//<


//> IF_INFORMATION_RB
//c This RB is returned by most of the interface configuration commands.
//c The {\tt ifno} field is set to {\tt -1} on error.
struct IF_INFORMATION_RB : public VRRB
{
  char ifno;
  char name[10];
  unsigned long ip;
  unsigned long netmask;
  unsigned long broadcast;
  unsigned long rx;
  unsigned long tx;
  unsigned long errors;
  unsigned long drops;
  unsigned long bandwidth;
  unsigned long bucketsize;
  unsigned long bucketlevel;
  unsigned long iptrans_tx_ip;
  unsigned long iptrans_tx_nm;
  unsigned long iptrans_tx_value;
  unsigned long iptrans_tx_pattern;
  unsigned long iptrans_rx_ip;
  unsigned long iptrans_rx_nm;
  unsigned long iptrans_rx_value;
  unsigned long iptrans_rx_pattern;
  char ctype;			//- 0: not connected\\ 1: softlink\\ 2: udp\\ 3: ipc\\ 4: ipip tunnel

  unsigned long cval[3];     	//- connection parameters
};
//<

#endif
