#include <stdlib.h>
#include <string.h>
#include "config.H"
#include "api.H"
#include "forwarder.H"
#include "tools.H"
#include "defragmenter.H"


Defragmenter::Defragmenter()
{
  ht=(DFRG_HASHTE*)malloc(sizeof(DFRG_HASHTE)*DFRG_HASHTABLE_SIZE);
  memset(ht,0,sizeof(DFRG_HASHTE)*DFRG_HASHTABLE_SIZE);
  freefrag_first=NULL;
  free_frags=0;
  alloc_frags=0;
  alloc_pkts=0;

  outbuff=(PSPacket*)malloc(sizeof(PSPacket)+0xffff);

  return;
}

Defragmenter::~Defragmenter()
{
  //
  // Datenstruktur freigeben
  //
  free(outbuff);
  free(ht);
  return;
}

unsigned short Defragmenter::hash(Packet *p)
{
  unsigned long s, i;
  s=PKT_SOURCE(p); i=PKT_ID(p);
  return ((s&0xff) ^ ((s>>8)&0xff) ^ ((s>>16)&0xff) ^ ((s>>24)&0xff) ^ (i&0xff) ^ ((i>>8)&0xff));  
}

PSPacket* Defragmenter::defragment(Packet *p)
{
  // the packet defragmenter
  // ....

#ifdef DEBUG_DFRG
  fhprintf(vr->cfh, "defragmenter: got packet '%s' free_frags: %lu alloc_frags:%lu  alloc_frags %lu\n",pkt2char(p), free_frags, allow_frags, alloc_pkts);
  fhprintf(vr->cfh,"defragmenter: alloc_pkts: %lu\n",alloc_pkts);
#endif

  // first or only fragment of an larger packet
  // so we build the pseudoheader and return the outbuff;
  if(!PKT_FRAG(p)){
    outbuff->source=PKT_SOURCE(p);
    outbuff->dest=PKT_DEST(p);
    outbuff->zero=0;
    outbuff->proto=PKT_PROTO(p);
    outbuff->pktlen=PKT_LENGTH(p);
    outbuff->datalen=PKT_LENGTH(p)-PKT_IHL(p)*4;
    memcpy(&outbuff->pkt, p, outbuff->pktlen);
    outbuff->data=PKT_DATA(&outbuff->pkt);
    return outbuff;
  }
 
  unsigned short h=hash(p);

#ifdef DEBUG_DFRG
  fhprintf(vr->cfh, "defragmenter: is a fragment with hash %u\n",h);
#endif

  // packet is fragment

  DFRG_PKT *n;

  if(ht[h].pkt==NULL){  // first packet (and fragment) for this hash
#ifdef DEBUG_DFRG
    fhprintf(vr->cfh,"defragmenter: packet was first packet for hash\n");
#endif
    ht[h].pktcnt=1;
    alloc_pkts++;
    n=ht[h].last_pkt=ht[h].pkt=(DFRG_PKT*)malloc(sizeof(DFRG_PKT));
    n->datalen=0;
    n->source=PKT_SOURCE(p);
    n->id=PKT_ID(p);
    n->next_pkt=NULL;
    n->first_frag=NULL;
    
    add_frag_to_fragchain(n,p);

    return NULL;
  }
  
  // schon (evtl passende?) pkts da

  DFRG_PKT *oldest_packet=NULL;
  double oldest_access, act_time;
  oldest_access=act_time=gettimeasdouble();

  for(n=ht[h].pkt;n;n=n->next_pkt){
    if(PKT_SOURCE(p)==n->source && PKT_ID(p)==n->id){ // pass. pkt schon registriert
      add_frag_to_fragchain(n, p);                    // an best. frags anhaengen

#ifdef DEBUG_DFRG
      fhprintf(vr->cfh,"defragmenter: added to existing fragments\n");;
#endif

      if(n->fraglen==n->datalen){
#ifdef DEBUG_DFRG
	fhprintf(vr->cfh,"defragmenter: packet complete\n");
#endif
	concat_frags(n);


#ifdef DEBUG_DFRGSTAT
	fhprintf(vr->cfh,"defragmenter: free: frags %lu alloc: frags %lu pkts %lu \n",
		 free_frags, alloc_frags, alloc_pkts);
#endif
	return outbuff;
      }
      
      return NULL; // neues fragment angehaengt 
    }

    //
    // old packet management
    //


    if(n->last_access<oldest_access || n->first_frag==0)
      oldest_access=n->last_access;
    oldest_packet=n;    
  }

  // erstes Fragment von neuem Packet
  
  // evtl. schon alloc packetslots benuetzen (oldest_packet, oldest_access)
  
#ifdef DEBUG_DFRG
  fhputs("defragmenter: new packet\n", vr->cfh);
#endif
  
  if(oldest_packet && (oldest_access+5<act_time || oldest_packet->first_frag==NULL)){
    
#ifdef DEBUG_DFRG
    fhprintf(vr->cfh,"defragmenter: reusing old slot dt: %f\n",  act_time-oldest_access);
#endif
    
    n=oldest_packet;
    n->last_access=act_time;
    if(n->first_frag)free_fragchain(n->first_frag, n->last_frag, n->fragcnt);
    n->first_frag=n->last_frag=NULL;
    n->source=PKT_SOURCE(p);
    n->id=PKT_ID(p);
    n->datalen=0;
    
#ifdef DEBUG_DFRG
    fhputs("defragmenter: adding to old packetslot\n",vr->cfh);
#endif
    add_frag_to_fragchain(n,p);
    
    return NULL;
  }
  
  // kein alter packetslot, neues packet anhaengen
  
#ifdef DEBUG_DFRG
  fhprintf(vr->cfh,"defragmenter: new packet added to hash, pkts: %u\n",ht[h].pktcnt);
#endif
  
  ht[h].pktcnt++;
  n=ht[h].last_pkt;
  alloc_pkts++;
  n->next_pkt=(DFRG_PKT*)malloc(sizeof(DFRG_PKT));
  n=n->next_pkt;
  n->next_pkt=NULL;
  n->source=PKT_SOURCE(p);
  n->id=PKT_ID(p);
  n->datalen=0;
    
  add_frag_to_fragchain(n,p);
  
  return NULL;
}

int Defragmenter::concat_frags(DFRG_PKT *n)
{
  DFRG_FRAG *m;
  
  // pseaudo header fuellen und ip header kopieren

  outbuff->source=PKT_SOURCE(n->first_frag->data);
  outbuff->dest=PKT_DEST(n->first_frag->data);
  outbuff->proto=PKT_PROTO(n->first_frag->data);
  outbuff->zero=0;
  outbuff->datalen=n->datalen;
  outbuff->pktlen=n->datalen+PKT_HDRLEN(n->first_frag->data);
  memcpy(&(outbuff->pkt), n->first_frag->data,PKT_HDRLEN(n->first_frag->data));
  outbuff->data=PKT_DATA(&outbuff->pkt);
  
  for(m=n->first_frag;m;m=m->next_frag){
#ifdef DEBUG_DFRG
    *con << "defragmenter: " << m->data << "\n";
#endif
    memcpy(outbuff->data+PKT_FOFF(m->data),PKT_DATA(m->data),PKT_DATALEN(m->data));
  }

#ifdef DEBUG_DFRG
    fhputs("defragmenter: garbage collection\n", vr->cfh);
#endif

  free_fragchain(n->first_frag, n->last_frag, n->fragcnt);
  n->first_frag=n->last_frag=NULL;

  return 0;
}

int Defragmenter::add_frag_to_fragchain(DFRG_PKT *q, Packet *p)
{
  DFRG_FRAG *m,*n;
  unsigned short f=PKT_FOFF(p);


#ifdef DEBUG_DFRG
  *con << "defragmenter: adding frag to fragchein\n";
  *con << "defragmenter: fragcnt:" << q->fragcnt << " fraglen:" << q->fraglen << "\n";
  *con << "defragmenter: datalen:" << q->datalen << "\n";
#endif
  
  if(q->first_frag==NULL){   // first fragment for fragchain
    q->fragcnt=0;
    n=q->first_frag=q->last_frag=alloc_frag();
    q->first_frag->next_frag=NULL;
    q->fraglen=0;
    goto frag_added;
  }
  
  for(m=q->first_frag; m ;m=m->next_frag){
    if(PKT_FOFF(m->data) > f )break;
    if(PKT_FOFF(m->data) == f )return 0; // duplicate packet
  }

  // Anfang der Kette 

  if(m==q->first_frag){
    n=alloc_frag();
    n->next_frag=q->first_frag;
    q->first_frag=n;
    goto frag_added;
  }

  // Ende der Kette

  if(!m){
    q->last_frag->next_frag=alloc_frag();
    q->last_frag=q->last_frag->next_frag;
    q->last_frag->next_frag=NULL;
    n=q->last_frag;
    goto frag_added;
  }

  // mittendrin
  
  n=m->next_frag;
  m->next_frag=alloc_frag();
  m=m->next_frag;
  m->next_frag=n;
  n=m;

 frag_added:
  
  ++q->fragcnt;
  q->last_access=gettimeasdouble();
  q->fraglen+=PKT_DATALEN(p);
  if(!PKT_MF(p))q->datalen=PKT_FOFF(p)+PKT_DATALEN(p);
  memcpy(n->data, p, PKT_LENGTH(p));
  
  return 0;
}

int Defragmenter::free_frag(DFRG_FRAG *f)
{
  if(!f) return 1;
  free_frags++;
  f->next_frag=freefrag_first;
  freefrag_first=f;
  return 0;
}


DFRG_FRAG* Defragmenter::alloc_frag()
{
  if(freefrag_first){
    DFRG_FRAG *f;
    f=freefrag_first;
    freefrag_first=freefrag_first->next_frag;
    --free_frags;
    return f;
  }

  alloc_frags++;
  return (DFRG_FRAG*)malloc(sizeof(DFRG_FRAG));
}


int Defragmenter::free_fragchain(DFRG_FRAG *first, DFRG_FRAG *last, int n)
{
  if(!first || !last)return 1;

  free_frags+=n;

  last->next_frag=freefrag_first;
  freefrag_first=first;
  return 0;
}
