#include <stdlib.h>
#include <string.h>
#include "vr.H"
#include "api.H"
#include "tools.H"
#include "forwarder.H"
#include "interface.H"
#include "icmp.H"
#include "socks.H"



ICMP::ICMP(VR *v): ProtocolStack()
{
  vr=v;
  packet_out=(Packet*)malloc(256);
  
  return;
}

ICMP::~ICMP()
{
  free(packet_out);
  return;
}


// generates an ICMP message due to TTL exceed
int ICMP::ttl_exceeded(Interface *ifc, Packet *in)
{
  // constant header part

  memset(packet_out,0,256);
  packet_out->version=4;
  packet_out->ttl=64;
  packet_out->ihl=5;
  packet_out->prot=1;
  packet_out->source=ifc->get_ip();
  packet_out->dest=in->source;
  packet_out->id=vr->pf->get_packet_id();

  // fill ICMP message
  //   0                   1                   2                   3
  //  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
  // +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  // |     Type      |     Code      |          Checksum             |
  // +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  // |                             unused                            |
  // +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  // |      Internet Header + 64 bits of Original Data Datagram      |
  // +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

  struct icmp_header *pic =(struct icmp_header*)(PKT_DATA(packet_out));
  pic->type=11;
  pic->code=0;
  pic->cksum=0;
  pic->id=0; 
  pic->sn=0;
  memcpy(&pic->data,in, PKT_HDRLEN(in)+8);
  pic->cksum=cksum((char*)pic,
		   sizeof(struct icmp_header)-1+
		   PKT_HDRLEN(in)+8);

  // fill rest of packet's header

  packet_out->len=htons(PKT_HDRLEN(packet_out)+
			sizeof(struct icmp_header)-1+
			PKT_HDRLEN(in)+8);

#ifdef DEBUG_ICMP_TTL
  fhprintf(vr->cfh, "--- DEBUG_ICMP --- (1)\n");
  fhprintf(vr->cfh, "ttl exceed message sent to %s\n", n2char(PKT_DEST(packet_out)));
#endif

  vr->pf->distribute(NULL, packet_out);
  return 0;
}



/* generates an ICMP message due to unreachable
  reason can be
  0 = net unreachable;
  1 = host unreachable;
  2 = protocol unreachable;
  3 = port unreachable;
  4 = fragmentation needed and DF set;
  5 = source route failed.
*/

int ICMP::unreachable(Interface *ifc, Packet *in, int reason)
{
  return unreachable(ifc->get_ip(), in, reason);
}


int ICMP::unreachable(unsigned long address, Packet *in, int reason)
{
  // constant header part

  memset(packet_out,0,256);
  packet_out->version=4;
  packet_out->ttl=64;
  packet_out->ihl=5;
  packet_out->prot=1;
  packet_out->source=address;
  packet_out->dest=in->source;
  packet_out->id=vr->pf->get_packet_id(); // neue id vegeben

  // fill ICMP message
  //   0                   1                   2                   3
  //  0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
  // +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  // |     Type      |     Code      |          Checksum             |
  // +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  // |                             unused                            |
  // +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  // |      Internet Header + 64 bits of Original Data Datagram      |
  // +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

  struct icmp_header *pic =(struct icmp_header*)(PKT_DATA(packet_out));
  pic->type=3;
  pic->code=reason;
  pic->cksum=0;
  memcpy(&pic->data,in, PKT_HDRLEN(in)+8);
  pic->cksum=cksum((char*)pic,
		   sizeof(struct icmp_header)-1+
		   PKT_HDRLEN(in)+8);

  // fill rest of packet's header

  packet_out->len=htons(PKT_HDRLEN(packet_out)+
			sizeof(struct icmp_header)-1+
			PKT_HDRLEN(in)+8);

#ifdef DEBUG_ICMP_UNREACH
  fhprintf(vr->cfh, "--- DEBUG_ICMP --- (2)\n");
  fhprintf(vr->cfh, "unreachable message (%d) sent to %s\n", 
	   reason, n2char( PKT_DEST(packet_out)));
#endif





  vr->pf->distribute(NULL, packet_out);
  return 0;
}


int ICMP::process(PSPacket *p)
{

#ifdef DEBUG_ICMP
  fhprintf(vr->cfh,"--- DEBUG_ICMP --- (3)\n");
  fhprintf(vr->cfh,"processing incoming ICMP packet\n");
  fhprintf(vr->cfh,"%s\n",pkt2char(&p->pkt)); 
#endif

  struct icmp_header *pic =(struct icmp_header*)p->data;
  
  switch(pic->type)
    {
    case 0:			// echo reply
#ifdef DEBUG_ICMP_ECHO
      fhprintf(vr->cfh,"--- DEBUG_ICMP --- (4)\n");
      fhprintf(vr->cfh,"packet contains echo reply -> raw socket dispatcher\n");
#endif
      vr->pf->raw_socks->dispatch(&p->pkt);
      break;

    case 8:  // echo request
#ifdef DEBUG_ICMP_ECHO
      fhprintf(vr->cfh,"--- DEBUG_ICMP --- (4)\n");
      fhprintf(vr->cfh,"packet contains echo request:\n");
      fhprintf(vr->cfh,"from %s id=%d sn=%d\n",n2char(p->source), pic->id, pic->sn);
#endif
      memset(packet_out,0,256);
      packet_out->version=4;
      packet_out->ttl=vr->ttl_default;
      packet_out->ihl=5;
      packet_out->prot=1;
      packet_out->source=p->dest;
      packet_out->dest  =p->source;
      packet_out->len=htons(84);
      packet_out->id=vr->pf->get_packet_id();

      pic->type=pic->code=0;
      pic->cksum=0;
      pic->cksum=cksum((char*)pic,64);
      memcpy(PKT_DATA(packet_out),pic,64);
      
      vr->pf->distribute(NULL, packet_out);
      break;

    default:

#ifdef DEBUG_ICMP
      fhprintf(vr->cfh,"--- DEBUG_ICMP --- (5)\n");
      fhprintf(vr->cfh,"unknown type -> raw socket dispatcher\n");
      fhprintf(vr->cfh,"type=%d sn=%d id=%d\n",pic->type, pic->sn, pic->id);
#endif
      packet_out=packet_out;
      vr->pf->raw_socks->dispatch(&p->pkt);
    }
  return 0;
}


