#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "qcmem.H"
#include "qcdroptail.H"
#include "qco.H"
#include "tools.H"

QCdroptail::QCdroptail(QCroot *q, int n) : QCO(q)
{
  ext_ql=0;
  max_nof_ins=1;  
  max_nof_outs=1;
  en_idx=0;
  de_idx=0;
  nof_pkts=0;
  max_nof_pkts=vr->dpt_queuelen;
  pkts=new (Packet*)[max_nof_pkts];
  ins=new QCO*[max_nof_ins];
  outs=new QCO*[max_nof_outs];
  ins[0]=NULL; 
  outs[0]=NULL;
  drops=0; 
  deempty=0;
  dehits=0; 
  dereq=0;
  return;
}

QCdroptail::~QCdroptail()
{
  delete pkts;
  delete ins;
  delete outs;
  return;
}


Packet * QCdroptail::deque(QCO *)
{
   
  if(!ins[0])return NULL;

#ifdef DEBUG_QUEUEING
  fhputs("--- deque ---\n",vr->cfh);
  fhprintf(vr->cfh,"%s (%d)\n", id(), get_id(this));
#endif

  Packet *p;
  p=ins[0]->deque(this);
  if(p)enque(p);
  
  if(nof_pkts){
    p=pkts[de_idx]; 
    pkts[de_idx]=NULL;
    ++de_idx;

    if(de_idx>=max_nof_pkts)de_idx=0;
    --nof_pkts;
    ++dehits;

    return p;
  }
  ++deempty;
  return NULL;
}

Packet *QCdroptail::nextdeque(QCO *x)
{
  if(!ins[0])return NULL;

  Packet *p; 
  
  p=ins[0]->deque(this);
  if(p)enque(p);
  
  if(nof_pkts>0){
    return pkts[de_idx];
  }
  return NULL;
}


int QCdroptail::enque(Packet *p)
{
  
#ifdef DEBUG_QUEUEING
  fhputs("--- enque ---",vr->cfh);
  fhprintf(vr->cfh,"%s (%d)\n", id(), get_id(this));
#endif

  if(p==NULL)
    fhprintf(vr->cfh,"%s (%d) enque of NULL packet !!!", id(), get_id(this));

  if(nof_pkts>=max_nof_pkts-1 || extended_drop_decision(p)){
    if(p==NULL)
      fhprintf(vr->cfh,"drop of null Packet!!\n");

    drops++; 
    drop(p);
    return QUEUE_ACK;
  }




  pkts[en_idx]=p;
  ++nof_pkts;
  ++en_idx;
  if(en_idx>=max_nof_pkts)en_idx=0;

  if(outs[0]){
    if(outs[0]->enque(pkts[de_idx])==QUEUE_ACK){
      deque(this);
    }
  }
  return QUEUE_ACK;
}
  

VRRB* QCdroptail::APIconfigure(VRCB *control_block)
{
  IF_QS_VRCB *cb=(IF_QS_VRCB*)control_block;

  if(cb->qcmd!=10){		// should be component specific

#ifdef DEBUG_QUEUEING
    fhputs("QCdroptail got wrong control block", vr->cfh); 
#endif
    return NULL;
  }
  
  switch(cb->component_command)
    {
    case 1:			// reset counters
      {
	drops=0; 
	deempty=0;
	dehits=0; 
	dereq=0;
	VRRB *rb=(VRRB*)malloc(sizeof(VRRB*));
	rb->len=sizeof(VRRB);
	rb->handle=cb->handle;
	return rb;
      }
    case 2:			// query statistics
      {
	QS_DROPTAIL_STATISTICS_VRRB *rb=
	  (QS_DROPTAIL_STATISTICS_VRRB*)malloc(sizeof(QS_DROPTAIL_STATISTICS_VRRB));
	rb->handle=cb->handle;
	rb->len=sizeof(QS_DROPTAIL_STATISTICS_VRRB);
	rb->component_type=1;	// droptail
	rb->component_id=get_id(this);
	rb->queuelen=max_nof_pkts;
	rb->dropped_packets=drops;
	rb->deque_requests=dereq;
	rb->deque_hits=dehits;
	rb->deque_fails=deempty;
	return rb;
      }
    case 310:			// set queuelen (not fully implemented)
      {
	QS_DROPTAIL_STATISTICS_VRRB *rb=
	  (QS_DROPTAIL_STATISTICS_VRRB*)malloc(sizeof(QS_DROPTAIL_STATISTICS_VRRB));
	rb->handle=cb->handle;
	rb->len=sizeof(QS_DROPTAIL_STATISTICS_VRRB);
	rb->component_type=1;	// droptail
	rb->component_id=get_id(this);
	rb->queuelen=max_nof_pkts;
	rb->dropped_packets=drops;
	rb->deque_requests=dereq;
	rb->deque_hits=dehits;
	rb->deque_fails=deempty;
	return rb;
      }
    }
  return NULL;
}


int QCdroptail::extended_drop_decision(Packet *p)
{
   return 0;			// no packet drop

  double w=0.02;

  ext_ql=(1.0-w)*ext_ql+w*nof_pkts; 


  if(ext_ql<0.5*max_nof_pkts)return 0; // no congestion
  
  struct vheader{
    unsigned short sp;
    unsigned short dp;
    unsigned short len;
    unsigned short crc; 
    unsigned short cookie;
    unsigned short pid;
    unsigned short foff;   
    unsigned short plen;
    unsigned short check;
  };
  
  if(PKT_PROTO(p)!=17)return 0;
  
  if(PKT_SOURCE(p)!=char2n("130.92.70.41"))return 0;
  if(PKT_DEST(p)!=char2n("10.42.14.2"))return 0;

  struct vheader *h=(struct vheader*)(PKT_DATA(p));

  if(h->cookie!=0x4242) return 0;
  if(h->pid%2)return 1;		// drop the packet
  return 0;			// leave it
}


