#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "qcdsmarker.H"
#include "tools.H"
#include "ipnmx.H"
#include "qco.H"

QCdsmarker::QCdsmarker(QCroot *q) : QCO(q)
{
  max_nof_ins=1; 
  max_nof_outs=1;
  ins=new QCO*[max_nof_ins];
  outs=new QCO*[max_nof_outs];
  ins[0]=NULL; 
  outs[0]=NULL;
  tbl=new IPNMx(1024);
  nof_rules=0;

  enques=0;
  newmarks=0;
  oldmarks=0;
  return;
}

QCdsmarker::~QCdsmarker() 
{
  delete ins;
  delete outs;
  delete tbl;
  return;
}

int QCdsmarker::enque(Packet *p)
{
#ifdef DEBUG_QUEUEING
  fhputs("--- enque ---",vr->cfh);
  fhprintf(vr->cfh,"%s (%d)\n", id(), get_id(this));
#endif

  if(!outs[0])return QUEUE_NAK;
  mark(p);
  return outs[0]->enque(p);
}


Packet *QCdsmarker::nextdeque(QCO*x)
{
  if(!ins[0])return 0;
  return ins[0]->nextdeque(this);
}


Packet * QCdsmarker::deque(QCO *x)
{

#ifdef DEBUG_QUEUEING
  fhputs("--- deque ---\n",vr->cfh);
  fhprintf(vr->cfh,"%s (%d)\n", id(), get_id(this));
#endif

  if(!ins[0])return NULL;
  Packet *p=ins[0]->deque(this);
  mark(p);
  return p;
}


int QCdsmarker::mark(Packet *p)
{
  if(!p)return 0;
  IPNMxRecord *r=tbl->lookup(p);

  if(!r){
    return 0;
  }

  // packet matches rule 

  DSstate_Info *si=(DSstate_Info*)r->content;

  if(si->service==100){
    p->tos=p->tos&3;
    p->tos=p->tos|184;
    newmarks++;
    // fhprintf(vr->cfh,"dsm: packet marked EF %s\n",pkt2char(p));
    return 100;
  }


  char pcd[3];
  switch(si->service)
    {
    case 201: pcd[0]=40;  pcd[1]=48;  pcd[2]=56;  break;
    case 202: pcd[0]=72;  pcd[1]=80;  pcd[2]=88;  break;
    case 203: pcd[0]=104; pcd[1]=112; pcd[2]=120; break;
    case 204: pcd[0]=136; pcd[1]=144; pcd[2]=152; break;
    }


  double ti=gettimeasdouble();
  for(int i=0;i<2;++i){
    si->bucket_level[i]+=(unsigned long)((ti-si->time_mark)*si->bw[i]);
    if(si->bucket_level[i]>si->bs[i])si->bucket_level[i]=si->bs[i];
  }
  si->time_mark=ti;

  unsigned int l=PKT_LENGTH(p);

  p->tos=p->tos&3;

  if(si->bucket_level[0]>=l){	// fits in low dprec bucket
    p->tos=p->tos|pcd[0];
    // fhprintf(vr->cfh,"dsm: packet AF low:  %s\n",pkt2char(p));
    si->bucket_level[0]-=l;
    return 1;
  }

  if(si->bucket_level[1]>=l){	// fits in medium dprec bucket
    p->tos=p->tos|pcd[1];
    // fhprintf(vr->cfh,"dsm: packet AF med: %s\n",pkt2char(p));
    si->bucket_level[1]-=l;
    return 1;
  }

  p->tos=p->tos|pcd[2];		// no bucket, high dprec  
  // fhprintf(vr->cfh,"dsm: packet AG hig: %s\n",pkt2char(p));
  return 1;
}



VRRB* QCdsmarker::APIconfigure(VRCB *control_block)
{
  IF_QS_VRCB *cb=(IF_QS_VRCB*)control_block;

#ifdef DEBUG_API
	fhprintf(vr->cfh,"DEBUG_API: DSmarker API handler (command=%d)\n",cb->component_command);
#endif

  if(cb->qcmd!=10){		// should be component specific
#ifdef DEBUG_API
    fhputs("QCdsmarker got wrong control block", vr->cfh); 
#endif
    return NULL;
  }
  
  switch(cb->component_command)
    {
    case 1:			// reset counters
      {
#ifdef DEBUG_API
	fhputs("QCdsmarker reset\n", vr->cfh); 
#endif
	enques=0;
	newmarks=0;
	oldmarks=0;

	VRRB *rb=(VRRB*)malloc(sizeof(VRRB*));
	rb->len=sizeof(VRRB);
	rb->handle=cb->handle;
	return rb;
      }
    case 2:			// query statistics
      {
#ifdef DEBUG_API
	fhputs("QCdsmarker query statistics\n", vr->cfh); 
#endif
	QS_DSMARKER_STATISTICS_VRRB *rb=
	  (QS_DSMARKER_STATISTICS_VRRB*)malloc(sizeof(QS_DSMARKER_STATISTICS_VRRB)+
					       nof_rules*sizeof(DS_MARKER_RECORD));
	rb->handle=cb->handle;
	rb->len=sizeof(QS_TBF_STATISTICS_VRRB);
	rb->component_type=type_id();
	rb->component_id=get_id(this);
	rb->enques=enques;
	rb->newmarks=newmarks;
	rb->oldmarks=oldmarks;

	DS_MARKER_RECORD *rec=(DS_MARKER_RECORD *)(rb+1); // append to VRRB 
	
	IPNMxRecord a;
	int r=tbl->get_first(a);
	int i=0;
	while(r){
	  rec[i].source_ip=a.sip_add;
	  rec[i].source_nm=a.snm_add;
	  rec[i].dest_ip=a.dip_add;	
	  rec[i].dest_nm=a.dnm_add;
	  rec[i].tos=a.tos;		
	  rec[i].proto=a.proto;	
	  struct DSstate_Info *si=(struct DSstate_Info*)(a.content);
	  rec[i].service=si->service;
	  rec[i].param[0]=si->bw[0];
	  rec[i].param[1]=si->bs[0];
	  rec[i].param[2]=si->bw[1];
	  rec[i].param[3]=si->bs[1];
	  ++i;
	  r=tbl->get_next(a);
	}
	rb->len=sizeof(QS_TRIO_STATISTICS_VRRB)+i*sizeof(DS_MARKER_RECORD);
	return (VRRB*)rb;
      }
      
    case 611: // add a DS Marker rule
      {
#ifdef DEBUG_API
	fhputs("QCdsmarker adding rule\n", vr->cfh); 
#endif
	struct dummy0 : public IF_QS_VRCB
	{
	  DS_MARKER_RECORD ds;
	};
	
	struct dummy1 : public QS_DSMARKER_STATISTICS_VRRB
	{
	  DS_MARKER_RECORD rec;
	};
	
	struct dummy0 *dscb=(struct dummy0*)cb;
	struct dummy1 *rb=(struct dummy1*)malloc(sizeof(struct dummy1));
	
	rb->handle=cb->handle;
	rb->len=sizeof(struct dummy1);
	rb->component_type=type_id(); 
	rb->component_id=get_id(this);

	memset(&(rb->rec),0,sizeof(DS_MARKER_RECORD));

	struct DSstate_Info *si=(struct DSstate_Info*)malloc(sizeof(struct DSstate_Info));
		
	if(!dscb->ds.param[1])dscb->ds.param[1]=8192;
	if(!dscb->ds.param[3])dscb->ds.param[3]=8192;

	si->service=dscb->ds.service;
	si->bw[0]=dscb->ds.param[0];
	si->bw[1]=dscb->ds.param[2];
	si->bs[0]=dscb->ds.param[1];
	si->bs[1]=dscb->ds.param[3];

	nof_rules++;
	
	IPNMxRecord a(dscb->ds.source_ip, dscb->ds.source_nm, 
		      dscb->ds.dest_ip, dscb->ds.dest_nm, 
		      dscb->ds.proto, dscb->ds.tos, (void*)si);
	tbl->add(a);
	
	memcpy(&rb->rec,&dscb->ds,sizeof(DS_MARKER_RECORD)); 
	
	return (VRRB*)rb;
      }

    case 612: // remove a DS Marker rule
      {
#ifdef DEBUG_API
	fhputs("QCdsmarker remove rule\n", vr->cfh); 
#endif
	struct dummy0 : public IF_QS_VRCB
	{
	  DS_MARKER_RECORD ds;
	};
	
	struct dummy1 : public QS_DSMARKER_STATISTICS_VRRB
	{
	  DS_MARKER_RECORD rec;
	};
	
	struct dummy0 *dscb=(struct dummy0*)cb;
	struct dummy1 *rb=(struct dummy1*)malloc(sizeof(struct dummy1));
	
	rb->handle=cb->handle;
	rb->len=sizeof(struct dummy1);
	rb->component_type=type_id(); 
	rb->component_id=get_id(this);
		
	IPNMxRecord a(dscb->ds.source_ip, dscb->ds.source_nm, dscb->ds.dest_ip, dscb->ds.dest_nm, 
		      dscb->ds.proto, dscb->ds.tos, NULL);

	if(!tbl->del(a)){	// success
	  nof_rules--;
	  memcpy(&rb->rec,&dscb->ds,sizeof(DS_MARKER_RECORD)); 
	  return (VRRB*)rb;
	}

	memset(&(rb->rec),0,sizeof(DS_MARKER_RECORD));
	rb->len=sizeof(QS_DSMARKER_STATISTICS_VRRB);
	return (VRRB*)rb;
      }
    }
  return NULL;
}



