#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "qcmem.H"
#include "qcroot.H"
#include "qcclassifier.H"
#include "qcdroptail.H"
#include "qcscheduler.H"
#include "qcdsmarker.H"
#include "qctbf.H"
#include "qctrio.H"
#include "tools.H"
#include "qco.H"
#include "config.H"

QCroot::QCroot(VR *v) : QCO(this)
{
  vr=v;
  qr=this;
  max_nof_qcos=256;
  qcos=new (QCO*)[max_nof_qcos];
  qcmem=new QCmem(512);
  
  max_nof_ins=1; 
  max_nof_outs=1;

  ins=new QCO*[max_nof_ins];
  outs=new QCO*[max_nof_outs];
  ins[0]=NULL; 
  outs[0]=NULL;

  deque_buffer=(Packet*)malloc(0xffff);

  memset(qcos, 0 , sizeof(QCO*)*max_nof_qcos);
  qcos[0]=this; 
  nof_qcos=1;

  // statistics counters 

  packets_dropped=0;		// overall packet drop
  packets_in_qs=0;		// overall packets in qs
  
  enques=0;			// number of sucessfull enques
  deques=0;			// number of sucessfull deques
};

QCroot::~QCroot()
{
  delete ins;
  delete outs;
  delete qcmem;

  free(deque_buffer);

  for(int i=0;i<nof_qcos;++i){
    if(qcos[i] && qcos[i]!=this){
      delete qcos[i];
    }
  }
  delete qcos;
  return;
}


Packet* QCroot::deque(QCO *)
{
#ifdef DEBUG_QUEUEING
  fhprintf(vr->cfh,"--- deque ---\n");
  fhprintf(vr->cfh,"%s (%d)\n", id(), get_id(this));
#endif
  return NULL;
}

int QCroot::enque(Packet *p)
{
#ifdef DEBUG_QUEUEING
  fhprintf(vr->cfh,"--- enque ---\n");
  fhprintf(vr->cfh,"%s (%d)\n", id(), get_id(this));
#endif
  return QUEUE_NAK;
}


Packet* QCroot::root_deque()
{
  if(!ins[0])return NULL;
  Packet *p=ins[0]->deque(this);
  if(p==NULL)return NULL;
  memcpy(deque_buffer, p, PKT_LENGTH(p));
  qcmem->release(p);
  ++deques;
  qr->packets_in_qs--;
  return deque_buffer;
}

Packet* QCroot::root_nextdeque()
{
  if(!ins[0])return NULL;
  Packet *p=ins[0]->nextdeque(this);
  return p;
}

int QCroot::root_enque(Packet *p)
{
  if(!outs[0])return QUEUE_NAK;
  Packet *P=qcmem->store(p);
  ++enques; 
  qr->packets_in_qs++;
  return outs[0]->enque(P);
}


VRRB* QCroot::APIconfigure(VRCB *v)
{
  IF_Q_VRCB* qcb=(IF_Q_VRCB*)v;
  
#ifdef DEBUG_API
	fhprintf(vr->cfh,"Queueing System API: qcroot command=%d\n",qcb->qcmd); 
#endif

  switch(qcb->qcmd)
    {
    case 1:			// list all components
      {
	VRRB* rb=(VRRB*)malloc(sizeof(VRRB)+qr->nof_qcos*2*sizeof(unsigned short));
	int j=0;
	unsigned short *n=(unsigned short*)(rb+1);
	for(int i=0;i<qr->nof_qcos;++i){
	  if(qr->qcos[i]!=NULL){
	    n[j++]=i;
	    n[j++]=qr->qcos[i]->type_id();
	  }
	}
	rb->handle=v->handle;
	rb->len=sizeof(VRRB)+j*sizeof(unsigned short);
	return rb;
      }
    case 2:			// list all connections

      {
	VRRB* rb=(VRRB*)malloc(sizeof(VRRB)+1024);
	rb->handle=v->handle;
	unsigned short *n=(unsigned short*)(rb+1);
	int k=0;
	for(int j=0;j<qr->nof_qcos;++j){
	  QCO *q=qr->qcos[j];
	  if(q!=NULL){
	    for(int o=0;o<q->nof_outs;++o){
	      if(q->outs[o]!=NULL){
		n[k++]=j;	// start id
		n[k++]=q->type_id(); // start type
		n[k++]=get_id(q->outs[o]);  // end id
		n[k++]=q->outs[o]->type_id() ; // end type
	      }
	    }
	  }
	}
	rb->len=sizeof(VRRB)+k*sizeof(unsigned short);
	return rb;
      }
    case 3:			// create component
      {
	VRRB* rb=(VRRB*)malloc(sizeof(VRRB)+2*sizeof(unsigned short));
	unsigned short *n=(unsigned short*)(rb+1);
	n[0]=0xffff; n[1]=0;
	rb->handle=v->handle;
	rb->len=sizeof(VRRB)+2*sizeof(unsigned short);

	/*
	 * get new index
	 */ 

	int idx=nof_qcos;
	for(int i=0; i<nof_qcos;++i){
	  if(qcos[i]==NULL){
	    idx=i; break;
	  }
	}

#ifdef DEBUG_API
	fhprintf(vr->cfh,"Creating component: new id %d\n",idx); 
#endif
	/*
	 * Component Creation
	 */

	switch(*(unsigned short*)(qcb+1))
	  {
	  case QCTYPEID_DROPTAIL:		// droptail
	    qcos[idx]=new QCdroptail(this);
	    n[0]=idx; n[1]=qcos[idx]->type_id();
	    if(idx==nof_qcos)++nof_qcos;
	    return (VRRB*)rb;

	  case QCTYPEID_TBF:		// tbf
	    qcos[idx] = new QCtbf(this);
	    n[0]=idx; n[1]=qcos[idx]->type_id();
	    if(idx==nof_qcos)++nof_qcos;
	    return (VRRB*)rb;
	    
	  case QCTYPEID_CLASSIFIER:		// classifier
	    qcos[idx]=new QCclassifier(this);
	    n[0]=idx; n[1]=qcos[idx]->type_id();
	    if(idx==nof_qcos)++nof_qcos;
	    return (VRRB*)rb;

	  case QCTYPEID_SCHEDULER:		// scheduler
	    qcos[idx]=new QCscheduler(this);
	    n[0]=idx; n[1]=qcos[idx]->type_id();
	    if(idx==nof_qcos)++nof_qcos;
	    return (VRRB*)rb;
	    
	  case QCTYPEID_TRIO:
	    qcos[idx]=new QCtrio(this);
	    n[0]=idx; n[1]=qcos[idx]->type_id();
	    if(idx==nof_qcos)++nof_qcos;
	    return (VRRB*)rb;
    
	  case QCTYPEID_DSMARKER:		// Differentiated Services Marker
	    qcos[idx]=new QCdsmarker(this);
	    n[0]=idx; n[1]=qcos[idx]->type_id();
	    if(idx==nof_qcos)++nof_qcos;
	    return (VRRB*)rb;
	  }

#ifdef DEBUG_API
	fhprintf(vr->cfh,"Error creating component with type id %d\n",
		 *(unsigned short*)(qcb+1)); 
#endif
	n[0]=0xffff; n[1]=0;	// Error
	return rb;

      }
    case 4:			// remove component
      { 
	VRRB* rb=(VRRB*)malloc(sizeof(VRRB)+2*sizeof(unsigned short));
	unsigned short *n=(unsigned short*)(rb+1);
	unsigned short *m=(unsigned short*)(qcb+1);
	n[0]=0xffff; n[1]=0;
	rb->handle=v->handle;
	rb->len=sizeof(VRRB)+2*sizeof(unsigned short);
	
	if(m[0]==0){		// root component cannot be deleted
	  n[0]=0xffff; n[1]=1;
	  return (VRRB*)rb;
	}
	
	if(m[0]>=nof_qcos || qcos[m[0]]==NULL){ // no such component
	  n[0]=0xffff; n[1]=2;
	  return (VRRB*)rb;
	}
				// component busy ??
	for(int i=0;i<nof_qcos;++i){
	  if(i==m[0])continue;
	  if(qcos[i]!=NULL &&  qcos[i]->isusing(qcos[m[0]])){
	    n[0]=0xffff; n[1]=3;
	    return (VRRB*)rb;
	  }
	}
	n[0]=m[0]; n[1]=qcos[m[0]]->type_id();
	delete qcos[m[0]];
	qcos[m[0]]=NULL;
	return (VRRB*)rb;
      }
    case 5:			// connect two components
      {
	VRRB *rb=(VRRB*)malloc(sizeof(VRRB)+2*sizeof(unsigned short));
	rb->len=sizeof(VRRB)+2*sizeof(unsigned short);
	rb->handle=v->handle;
	unsigned short *n=(unsigned short*)(rb+1);      
	n[0]=0xffff; n[1]=0;
	if(v->len!=sizeof(IF_Q_VRCB)+2*sizeof(unsigned short)){
	  n[1]=5;
	  return (VRRB*)rb;
	}
	unsigned short *m=(unsigned short*)(qcb+1);      
	if(m[0]>=nof_qcos || qcos[m[0]]==NULL){
	  n[1]=1;
	  return (VRRB*)rb;
	}
	if(m[1]>=nof_qcos || qcos[m[1]]==NULL){
	  n[1]=2;
	  return (VRRB*)rb;
	}
	if(qcos[m[0]]->next(qcos[m[1]])){
	  n[1]=3;
	  return (VRRB*)rb;
	}
	if(qcos[m[1]]->prev(qcos[m[0]])) {
	  n[1]=4;
	  return (VRRB*)rb;
	}
	n[0]=m[0]; n[1]=m[1];
	return (VRRB*)rb;
      }
    case 6:			// disconnect two components
      {
	VRRB *rb=(VRRB*)malloc(sizeof(VRRB)+2*sizeof(unsigned short));
	rb->len=sizeof(VRRB)+2*sizeof(unsigned short);
	rb->handle=v->handle;
	unsigned short *n=(unsigned short*)(rb+1);      
	n[0]=0xffff; n[1]=0;
	if(v->len!=sizeof(IF_Q_VRCB)+2*sizeof(unsigned short)){
	  n[1]=5;
	  return (VRRB*)rb;
	}
	unsigned short *m=(unsigned short*)(qcb+1);      
	if(m[0]>=nof_qcos || qcos[m[0]]==NULL){
	  n[1]=1;
	  return (VRRB*)rb;
	}
	if(m[1]>=nof_qcos || qcos[m[1]]==NULL){
	  n[1]=2;
	  return (VRRB*)rb;
	}
	if(qcos[m[0]]->unnext(qcos[m[1]])){
	  n[1]=3;
	  return (VRRB*)rb;
	}
	if(qcos[m[1]]->unprev(qcos[m[0]])) {
	  n[1]=4;
	  return (VRRB*)rb;
	}

	n[0]=m[0]; n[1]=m[1];
	return (VRRB*)rb;
      }
    case 9:			// add new component type
      break;
    case 10:			// commands specific for a component
      {
	IF_QS_VRCB *cb=(IF_QS_VRCB*)v;
	if(cb->component_id==0){		// it's me !!!
	  switch(cb->component_command){
	  case 1:		// reset counters
	    {
	      VRRB *rb=(VRRB*)malloc(sizeof(VRRB));
	      rb->handle=v->handle;
	      rb->len=sizeof(VRRB);
	      enques=0;
	      deques=0;
	      return rb;
	    }
	  case 2:		// return statistics
	    {
	      QS_ROOT_STATISTICS_VRRB *rb=
		(QS_ROOT_STATISTICS_VRRB*)malloc(sizeof(QS_ROOT_STATISTICS_VRRB));
	      rb->len=sizeof(QS_ROOT_STATISTICS_VRRB);
	      rb->handle=cb->handle;
	      rb->component_type=0; // root component
	      rb->component_id=0;   // root component
	      rb->enques=enques;
	      rb->deques=deques;
	      rb->global_stored=packets_in_qs;
	      rb->global_dropped=packets_dropped;
	      return (VRRB*)rb;
	    }
	  }
	  VRRB *rb=(VRRB*)malloc(sizeof(VRRB));
	  rb->handle=v->handle;
	  rb->len=sizeof(VRRB);
	  return rb;
	}
	
	if(qcos[cb->component_id]==NULL)break; // error
	
				// forward request component

#ifdef DEBUG_API
	fhprintf(vr->cfh,"forwarding to interface API handlers\n");
#endif

	VRRB *rb=qcos[cb->component_id]->APIconfigure(v);
	if(rb!=NULL)return rb;
      }
    }

  // something went wrong, return empty RB

  VRRB *rb=(VRRB*)malloc(sizeof(VRRB));
  rb->handle=v->handle;
  rb->len=sizeof(VRRB);
  return rb;
}


