#include <stdio.h>
#include <stdlib.h>
#include "qctbf.H"
#include "tools.H"
#include "qco.H"


QCtbf::QCtbf(QCroot *q) : QCO(q)
{
  bandwidth=vr->tbf_bandwidth;
  bucket_size=vr->tbf_bucketsize;
  bucket_level=0;

  last_call=gettimeasdouble();
  max_nof_ins=1; max_nof_outs=1;
  ins=new QCO*[max_nof_ins];
  outs=new QCO*[max_nof_outs];
  ins[0]=NULL; 
  outs[0]=NULL;
  bucket_overflows=0;
  enfail=0;
  enhits=0;
  enempty=0;
  defail=0;
  dehits=0;
  deempty=0;
}

QCtbf::~QCtbf() 
{
  delete ins;
  delete outs;
  return;
}

int QCtbf::enque(Packet *p)
{

#ifdef DEBUG_QUEUEING
  fhputs("--- enque ---",vr->cfh);
  fhprintf(vr->cfh,"%s (%d)\n", id(), get_id(this));
#endif

  if(!outs[0])return QUEUE_NAK;

  double ti=gettimeasdouble();
  bucket_level+=(unsigned long)((ti-last_call)*(double)bandwidth);
  last_call=ti;
  if(bucket_level>bucket_size){
    bucket_overflows++;
    bucket_level=bucket_size;
    }
  if(bucket_level<PKT_LENGTH(p)) {
    ++enfail;
    return QUEUE_NAK;
  }
  if(outs[0]->enque(p)==QUEUE_ACK){
    bucket_level-=PKT_LENGTH(p);
    ++enhits;
    return QUEUE_ACK;
  }
  enempty++;
  return QUEUE_NAK;
}


Packet *QCtbf::nextdeque(QCO*x)
{
  if(!ins[0])return 0;


  Packet *p=ins[0]->nextdeque(this);
  if(!p)return NULL;		// no packet to deque

  double ti=gettimeasdouble();
  bucket_level+=(unsigned long)((ti-last_call)*(double)bandwidth);
  last_call=ti;
  if(bucket_level>bucket_size)bucket_level=bucket_size;

  if(bucket_level>=PKT_LENGTH(p)){
    return p;
  }
  return NULL;
}


Packet * QCtbf::deque(QCO *x)
{

#ifdef DEBUG_QUEUEING
  fhputs("--- deque ---\n",vr->cfh);
  fhprintf(vr->cfh,"%s (%d)\n", id(), get_id(this));
#endif

  if(!ins[0])return NULL;

  Packet *p=ins[0]->nextdeque(this);
  if(!p){
    ++deempty;
    return NULL;
  }
  
  double ti=gettimeasdouble();

  bucket_level+=(unsigned long)((ti-last_call)*(double)bandwidth);
  last_call=ti;
  if(bucket_level>bucket_size)bucket_level=bucket_size;


  if(bucket_level>=PKT_LENGTH(p)){
    bucket_level-=PKT_LENGTH(p);
    ++dehits;
    return ins[0]->deque(this);
  }
  ++defail;

  return NULL;
}

VRRB* QCtbf::APIconfigure(VRCB *control_block)
{
  IF_QS_VRCB *cb=(IF_QS_VRCB*)control_block;

  if(cb->qcmd!=10){		// should be component specific

#ifdef DEBUG_QUEUEING
    fhputs("QCtbf got wrong control block", vr->cfh); 
#endif
    return NULL;
  }
  
  switch(cb->component_command)
    {
    case 1:			// reset counters
      {
	enhits=0;
	enfail=0;
	enempty=0;
	dehits=0;
	defail=0;
	deempty=0;
	bucket_overflows=0;
	VRRB *rb=(VRRB*)malloc(sizeof(VRRB*));
	rb->len=sizeof(VRRB);
	rb->handle=cb->handle;
	return rb;
      }
    case 2:			// query statistics
      {
	QS_TBF_STATISTICS_VRRB *rb=
	  (QS_TBF_STATISTICS_VRRB*)malloc(sizeof(QS_TBF_STATISTICS_VRRB));
	rb->handle=cb->handle;
	rb->len=sizeof(QS_TBF_STATISTICS_VRRB);
	rb->component_type=2;	// droptail
	rb->component_id=get_id(this);
	rb->enque_hits=enhits;
	rb->enque_fails=enfail;
	rb->enque_empty=enempty;
	rb->deque_hits=dehits;
	rb->deque_fails=defail;
	rb->deque_empty=deempty;
	rb->bucket_overflows=bucket_overflows;
	rb->bandwidth=bandwidth*8;
	rb->bucketsize=bucket_size*8;
	return rb;
      }
      
    case 210: // configure and reply statistics
      {
	struct dummy0 : public IF_QS_VRCB 
	{
	  unsigned long bw;
	  unsigned long bs;
	} *dcb;
	
	dcb=(struct dummy0*)cb;
	if(dcb->bw!=0xffffffff)bandwidth=dcb->bw/8;
	if(dcb->bs!=0xffffffff)bucket_size=dcb->bs/8;

	QS_TBF_STATISTICS_VRRB *rb=
	  (QS_TBF_STATISTICS_VRRB*)malloc(sizeof(QS_TBF_STATISTICS_VRRB));
	rb->handle=cb->handle;
	rb->len=sizeof(QS_TBF_STATISTICS_VRRB);
	rb->component_type=2;	// tbf
	rb->component_id=get_id(this);
	rb->enque_hits=enhits;
	rb->enque_fails=enfail;
	rb->enque_empty=enempty;
	rb->deque_hits=dehits;
	rb->deque_fails=defail;
	rb->deque_empty=deempty;
	rb->bucket_overflows=bucket_overflows;
	rb->bandwidth=bandwidth*8;
	rb->bucketsize=bucket_size*8;
	return rb;
      }
    
    }
  return NULL;
}




