#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "qcmem.H"
#include "qctrio.H"
#include "qco.H"
#include "tools.H"

QCtrio::QCtrio(QCroot *q, int n) : QCO(q)
{
  max_nof_ins=1;  
  max_nof_outs=1;
  en_idx=0;
  de_idx=0;
  nof_pkts=0;
  max_nof_pkts=vr->dpt_queuelen;
  pkts=new (Packet*)[max_nof_pkts];
  ins=new QCO*[max_nof_ins];
  outs=new QCO*[max_nof_outs];
  ins[0]=NULL; 
  outs[0]=NULL;

  queuelen[0]=(int)(0.33*max_nof_pkts);
  queuelen[1]=(int)(0.66*max_nof_pkts);
  queuelen[2]=max_nof_pkts-1;
  memset(dropped_packets,0,sizeof(unsigned long)*3);

  mode=0; // hard
  deempty=0;
  dehits=0; 
  dereq=0;
  return;
}


int QCtrio::resize(int len)
{
  Packet **np;
  np=new (Packet*)[len];

  // empty queue

  for(int i=0;i<max_nof_pkts;++i)
    if(pkts[i]){nof_pkts--; drop(pkts[i]);}

  delete pkts;
  pkts=np;
  en_idx=0;
  de_idx=0;
  max_nof_pkts=len;
  return 0;
}

QCtrio::~QCtrio()
{
  delete pkts;
  delete ins;
  delete outs;
  return;
}


Packet * QCtrio::deque(QCO *)
{
   
  if(!ins[0])return NULL;

#ifdef DEBUG_QUEUEING
  fhputs("--- deque ---\n",vr->cfh);
  fhprintf(vr->cfh,"%s (%d)\n", id(), get_id(this));
#endif

  Packet *p;
  p=ins[0]->deque(this);
  if(p)enque(p);
  
  if(nof_pkts){
    p=pkts[de_idx]; 
    pkts[de_idx]=NULL;
    ++de_idx;

    if(de_idx>=max_nof_pkts)de_idx=0;
    --nof_pkts;
    ++dehits;

    return p;
  }
  ++deempty;
  return NULL;
}

Packet *QCtrio::nextdeque(QCO *x)
{
  if(!ins[0])return NULL;

  Packet *p; 
  
  p=ins[0]->deque(this);
  if(p)enque(p);
  
  if(nof_pkts>0){
    return pkts[de_idx];
  }
  return NULL;
}


int QCtrio::enque(Packet *p)
{
  
#ifdef DEBUG_QUEUEING
  fhputs("--- enque ---",vr->cfh);
  fhprintf(vr->cfh,"%s (%d)\n", id(), get_id(this));
#endif

  if(p==NULL)
    fhprintf(vr->cfh,"PANIC: %s (%d) enque of NULL packet !!!", id(), get_id(this));

  unsigned short tos=PKT_TOS(p);
  
  // high drop precedence

  if((tos==0 || (tos&56)==56 || (tos&88)==88 || (tos&120)==120 || (tos&152)==152) && nof_pkts>=queuelen[0]){
    // fhprintf(vr->cfh,"high drop (%u): %s\n", nof_pkts, pkt2char(p));
    dropped_packets[0]++;
    drop(p);
    return QUEUE_ACK;
  }

  // medium drop precedence

  if(((tos&48)==48 || (tos&80)==80 || (tos&112)==112 || (tos&144)==144) && nof_pkts>=queuelen[1]){
    // fhprintf(vr->cfh,"medium drop (%u): %s\n", nof_pkts, pkt2char(p));
    dropped_packets[1]++;
    drop(p);
    return QUEUE_ACK;
  }

  // low drop precedence

  if(nof_pkts>=queuelen[2]){
    // fhprintf(vr->cfh,"low drop (%u): %s\n", nof_pkts, pkt2char(p));
    dropped_packets[2]++;
    drop(p);
    return QUEUE_ACK;
  }

  pkts[en_idx]=p;
  ++nof_pkts;
  ++en_idx;
  if(en_idx>=max_nof_pkts)en_idx=0;

  if(outs[0]){
    if(outs[0]->enque(pkts[de_idx])==QUEUE_ACK){
      deque(this);
    }
  }
  return QUEUE_ACK;
}
  

VRRB* QCtrio::APIconfigure(VRCB *control_block)
{
  IF_QS_VRCB *cb=(IF_QS_VRCB*)control_block;

  if(cb->qcmd!=10){		// should be component specific

#ifdef DEBUG_QUEUEING
    fhputs("QCtrio got wrong control block", vr->cfh); 
#endif
    return NULL;
  }
  
  switch(cb->component_command)
    {
    case 1:			// reset counters
      {
#ifdef DEBUG_API
    fhprintf(vr->cfh, "QCtrio: resetting counters\n");
#endif	

	memset(dropped_packets,0,sizeof(unsigned long)*3);
	deempty=0;
	dehits=0; 
	dereq=0;
	VRRB *rb=(VRRB*)malloc(sizeof(VRRB*));
	rb->len=sizeof(VRRB);
	rb->handle=cb->handle;
	return rb;
      }
    case 2:			// query statistics
      {
#ifdef DEBUG_API
    fhprintf(vr->cfh, "QCtrio: query statistics\n");
#endif	

	QS_TRIO_STATISTICS_VRRB *rb=
	  (QS_TRIO_STATISTICS_VRRB*)malloc(sizeof(QS_TRIO_STATISTICS_VRRB));
	rb->handle=cb->handle;
	rb->len=sizeof(QS_TRIO_STATISTICS_VRRB);
	rb->component_type=type_id();	// trio
	rb->component_id=get_id(this);
	memcpy(rb->queuelen,queuelen,sizeof(unsigned short)*3);
	memcpy(rb->dropped_packets,dropped_packets,sizeof(unsigned long)*3);
	rb->pkts=nof_pkts;
	rb->deque_requests=dereq;
	rb->deque_hits=dehits;
	rb->deque_fails=deempty;

	for(int i=0,j=de_idx; i<nof_pkts;++i){
	  if(j>=max_nof_pkts)j=0;
	  if(pkts[j]!=NULL)printf("%d %d %s\n",i,j, pkt2char(pkts[j]));
	  j++;
	}
	for(int i=0;i<3;++i){
	  printf("%d %d\n", i, queuelen[i]);
	}

	return rb;
      }
    case 710:			// set queuelen (not fully implemented)
      {
	struct dummy0 : public IF_QS_VRCB 
	{
	  unsigned short mode;
	  unsigned short queuelen[3];	 
	};
	
	struct dummy0 *lcb=(struct dummy0*)cb;

	if(lcb->queuelen[0]!=0xffff && lcb->queuelen[1]!=0xffff && lcb->queuelen[2]!=0xffff){
	  queuelen[0]=queuelen[2]*lcb->queuelen[0]/lcb->queuelen[2];
	  queuelen[1]=queuelen[2]*lcb->queuelen[1]/lcb->queuelen[2];
	}
	if(lcb->mode==0 || lcb->mode==2){
	  mode=lcb->mode;
	}

#ifdef DEBUG_API
    fhprintf(vr->cfh, "QCtrio: new queulengths: 0: %d, 1: %d, 2: %d new mode %d\n", 
	     queuelen[0], queuelen[1], queuelen[2],mode);
#endif	

	QS_TRIO_STATISTICS_VRRB *rb=
	  (QS_TRIO_STATISTICS_VRRB*)malloc(sizeof(QS_TRIO_STATISTICS_VRRB));
	rb->handle=cb->handle;
	rb->len=sizeof(QS_TRIO_STATISTICS_VRRB);
	rb->component_type=type_id();	// trio
	rb->component_id=get_id(this);
	rb->mode=mode;
	memcpy(rb->queuelen,queuelen,sizeof(unsigned short)*3);
	memcpy(rb->dropped_packets,dropped_packets,sizeof(unsigned long)*3);
	rb->deque_requests=dereq;
	rb->deque_hits=dehits;
	rb->deque_fails=deempty;
	return rb;
      }

    }
  return NULL;
}
