#include <stdio.h>
#include <limits.h>
#include <string.h>
#include <getopt.h>
#include <stdlib.h>
#include "vr.H"
#include "tools.H"


int show_help()
{

  fprintf(stderr,"Virtual Router\n");
  fprintf(stderr,"(c) Florian Baumgartner \n");
  fprintf(stderr,"url: http://www.iam.unibe.ch/~baumgart\n");
  fprintf(stderr,"\n");
  fprintf(stderr,"Syntax:\n");
  fprintf(stderr,"\t -f configfile\n");
  fprintf(stderr,"\t -s startupscript\n");
  fprintf(stderr,"\t -p portnumber (default is console)\n");
  fprintf(stderr,"\t -h hostname (default is local hostname)\n");
  return 0;
}

int get_prefix(char *a, char *s, int l)
{
  char buffer[PATH_MAX];	// complete path to binaryr
  realpath(a,buffer);
  if(strlen(buffer) < (strlen(a) + strlen("/bin/")))return -1;
  buffer[strlen(buffer)-strlen("router")-strlen("/bin/")]=0;
  memcpy(s,buffer,l);
  if((int)strlen(buffer)>=l)return -1;
  return 0;
}  


int main(int argc, char **argv)
{
 
  char prefix[512];
  char buffer[1024];
  VR* vr = new VR(); 

  /*
    for(int i=0;i<=32;++i)
    {
    printf("%3d : %8lx  %20s\n", i, n2nm(i), n2char(n2nm(i)));
    }
  */


  /* 
   * get the installation directory
   */

  if(get_prefix(argv[0],prefix,512)){
    puts("Error: cannot compute prefix\n");
    exit(0);
  }

  /*
   * adapt paths to standard values
   */

  sprintf(buffer,"%s/fifos/link",prefix);
  vr->set_fifobase(buffer);
  
  sprintf(buffer,"%s/lib",prefix);
  vr->set_lob_path(buffer);

  /*
   * parse command line parameters
   */

  while(1){
    int i=getopt(argc,argv,"f:s:p:h:");
    if(i<=0)break;
    switch(i)
      {
      case 'f':
	{
	  char *file=strdup(optarg);
	  vr->configurebyfile(file);
	  break;
	}
      case 's':	
	{
	  char *file=strdup(optarg);
	  vr->set_startupfile(file);
	  break;
	}
      case 'p':
	{
	  vr->set_portnumber(atoi(optarg));
	  break;
	}
      case 'h':
	{
	  vr->set_hostname(optarg);
	  break;
	}
      default:
	{
	  show_help();
	  exit(0);
	}
      }
  }

  vr->init();
  vr->run();
  
  return 0;
}




  









