#include <iostream.h>
#include <string.h>
#include <stdlib.h>
#include "tools.H"
#include "config.H"
#include "packet.H"
#include "interface.H"
#include "api.H"
#include "routing.H"


Routing::Routing(VR *v)
{
  vr=v;
  table=new RTable();
}

Routing::~Routing()
{
  delete(table);
  return;
}


VRRB* Routing::APIconfigure(VRCB *p) 
{
  ROUTE_RECORD_VRCB *cb=(ROUTE_RECORD_VRCB*)p;

  switch(cb->command)
    {
    case 1: // add a route
      {
	ROUTE_RECORD_VRRB *rb=(ROUTE_RECORD_VRRB*)malloc(sizeof(ROUTE_RECORD_VRRB));
	RTableRec *rt=new RTableRec(cb->dest_ip, cb->dest_nm, cb->ifid);
	int r=table->add(*rt);
	RTableRec2ROUTE_VRRB(rt,rb);
	delete rt;
	rb->len=sizeof(ROUTE_RECORD_VRRB);
	rb->handle=cb->handle;
	if(r)rb->ifid=0xffff;	// error
	return rb;
      }
    case 2: // delete a route
      {
	ROUTE_RECORD_VRRB *rb=(ROUTE_RECORD_VRRB*)malloc(sizeof(ROUTE_RECORD_VRRB));
	RTableRec *rt=new RTableRec(cb->dest_ip, cb->dest_nm, cb->ifid);
	int r=table->del(*rt);
	RTableRec2ROUTE_VRRB(rt,rb);
	delete rt;
	rb->len=sizeof(ROUTE_RECORD_VRRB);
	rb->handle=cb->handle;
	if(r)rb->ifid=0xffff;	// error
	return rb;
      }
    case 3: // list all routes
      {
	VRRB *rb=(VRRB*)malloc(sizeof(VRRB)+table->nof_entries*sizeof(ROUTE_RECORD));
	ROUTE_RECORD *rec=(ROUTE_RECORD*)(rb+1);
	for(int i=0;i<table->nof_entries;++i){
	  RTableRec2ROUTE_RECORD(table->table[i],&(rec[i]));
	}
	rb->len=sizeof(VRRB)+table->nof_entries*sizeof(ROUTE_RECORD);
	rb->handle=cb->handle;
	return rb;
      }
    }
  return NULL;
}

void Routing::RTableRec2ROUTE_VRRB(RTableRec *r, ROUTE_RECORD_VRRB *v)
{
  v->source_ip=r->sip_add;
  v->source_nm=r->snm_add;
  v->dest_ip=r->dip_add;
  v->dest_nm=r->dnm_add;
  v->proto=r->proto;
  v->tos=r->tos;
  v->ifid=r->content;
  v->len=sizeof(ROUTE_RECORD_VRRB);
  return;
}
void Routing::RTableRec2ROUTE_RECORD(RTableRec *r, ROUTE_RECORD *v)
{
  v->source_ip=r->sip_add;
  v->source_nm=r->snm_add;
  v->dest_ip=r->dip_add;
  v->dest_nm=r->dnm_add;
  v->proto=r->proto;
  v->tos=r->tos;
  v->ifid=r->content;
  return;
}

/*
 * loescht Eintraege, deren Interfaces es nicht mehr gibt
 */

void Routing::cleanup()
{
  // routen loeschen, deren Index es nicht mehr gibt

  for(int i=0;i<table->nof_entries;++i){
    if(!table->table[i])continue;
    if(vr->in[table->table[i]->content]==NULL){
      delete table->table[i];
      table->table[i]=NULL;
    }
  }

  // ueberfluesseige Einbtraege loeschen
   
  table->cleanup();
  return;
}


int Routing::route(Packet *p)
{
  
  RTableRec *r=table->lookup(p);
  if(r){
    if(vr->in[r->content]!=NULL){
#ifdef DEBUG_ROUTING
      fhprintf(vr->cfh,"DEBUG_ROT %s sent to %s\n",pkt2char(p),vr->in[r->content]->id());
#endif
      vr->in[r->content]->send_packet(p);
      return r->content;
    }
  }
  return 0xffff;		// error, no route found
}


Interface * Routing::if_lookup(Packet *p)
{
  RTableRec *r=table->lookup(p);
  if(r && vr->in[r->content])return vr->in[r->content];
  return NULL;			// no route found
}


RTable::RTable(int n)
{
  nof_allocated=n;
  nof_entries=0;
  table=new RTableRec*[nof_allocated];
  memset(table,0,sizeof(RTableRec*)*nof_allocated);
  return;
}

RTable::~RTable()
{
  for(int i=0;i<nof_allocated;++i){
    delete table[i];
  }
  return;
}

RTableRec* RTable::lookup(Packet *p)
{
  int h=0;
  RTableRec *v=NULL;

  for(int i=0;i<nof_entries;++i){
    int k=table[i]->hits(p);
    if(k>h){
      h=k; 
      v=table[i];
    }
  }
  return v;
}

int RTable::add(RTableRec& m)
{
  m.init();			// calculate bitmasks

  // check whether the entry already exists

  for(int i=0;i<nof_entries;++i)
    {
    if(table[i]->equals(&m))
      {
	return 1;
      }
    }
  
  if(nof_entries+1>nof_allocated)return 1;
  table[nof_entries]=new RTableRec(m);
  table[nof_entries]->init();
  nof_entries++;
  sort();
  return 0;
}


int RTable::del(RTableRec &m)
{
  int i;

  m.init();			// calculate bitmasks 

  for(i=0;i<nof_entries;++i)
    {
    if(table[i]->equals(&m))
      {
	delete table[i];
	table[i]=NULL;
	break;
      }
    }
  if(i==nof_entries) return 1;
  cleanup();
  sort();
  return 0;
}

void RTable::sort()
{
  int m;
  do{
    m=0;
    for(int i=0;i<nof_entries-1;++i){
      if(table[i+1]->gt(table[i])){
	RTableRec *r;
	r=table[i];
	table[i]=table[i+1];
	table[i+1]=r;
	++m;
      }
    }
  }while(m);
  return;
}

/*
 * entfernt geloeschte Eintraege. 
 */

void RTable::cleanup()
{
  int i,j;
  for(i=j=0;i<nof_entries;++i){
    if(table[i]!=NULL){
      table[j++]=table[i];
    }
  }
  nof_entries=j;
  return;
}

// ---------------------------------------------------------------------------

bool RTableRec::equals(RTableRec *r)
{
  return !(bool)bcmp(this,r,sizeof(RTableRec));
}

bool RTableRec::gt(RTableRec *r)
{
  if(this->dip_add > r->dip_add)return 1;
  if(this->dip_add < r->dip_add)return 0;
  if(this->dnm_num > r->dnm_num)return 1;
  if(this->dnm_num < r->dnm_num)return 0;

  if(this->sip_add > r->sip_add)return 1;
  if(this->sip_add < r->sip_add)return 0;
  if(this->snm_num > r->snm_num)return 1;
  if(this->snm_num < r->snm_num)return 0;

  if(this->proto > r->proto)return 1;
  if(this->proto < r->proto)return 0;

  if(this->tos > r->tos)return 1;
  if(this->tos < r->tos)return 0;

  return 0;
}


bool RTableRec::lt(RTableRec *r)
{
  return !gt(r);
}

int RTableRec::hits(Packet *p)
{
  int k=1;
  if(dip_add){
    if((PKT_DEST(p)&dnm_add) != (dip_add&dnm_add))return 0;
    k+=dnm_num+1;
  }
  if(sip_add){
    if((PKT_SOURCE(p)&snm_add) != (sip_add&snm_add))return 0;
    k+=snm_num+1;
  }
  if(proto){
    if(proto!=PKT_PROTO(p))return 0;
    k+=1;
  }
  if(tos){
    if(tos!=PKT_TOS(p))return 0;
    k+=1;
  }
  return k;
}


RTableRec::RTableRec(Packet *p)
{
  sip_add=PKT_SOURCE(p);
  snm_add=0xffffffff;
  dip_add=PKT_DEST(p);
  dnm_add=0xffffffff;
  proto=PKT_PROTO(p);
  tos=PKT_TOS(p);
  content=0xffffffff;
  return;
}

RTableRec::RTableRec(unsigned long ip, unsigned long nm, unsigned long v)
{
  memset(this,0,sizeof(RTableRec));
  dip_add=ip;
  dnm_add=nm;
  dnm_num=nm2n(nm);
  content=v;
  return;
}

int RTableRec::set(char *d, char *s, int p, int t, unsigned long v)
{
  memset(this,0,sizeof(RTableRec));
  char2ipnm(s, &sip_add, &snm_add);
  char2ipnm(d, &dip_add, &dnm_add);
  proto=p;
  tos=t;
  content=v;
  return 0;
}

RTableRec::RTableRec(char *d, char *s, int p, int t, unsigned long v)
{
  memset(this,0,sizeof(RTableRec));
  char2ipnm(s, &sip_add, &snm_add);
  char2ipnm(d, &dip_add, &dnm_add);
  proto=p;
  tos=t;
  content=v;
  return;
}


void RTableRec::clear()
{
  memset(this,0,sizeof(RTableRec));
  return;
}

void RTableRec::init()
{
  snm_num=nm2n(snm_add);
  dnm_num=nm2n(dnm_add);
  return;
}


RTableRec::RTableRec()
{
  sip_add=0; snm_add=0; snm_num=0;
  dip_add=0; dnm_add=0; dnm_num=0;
  proto=0; tos=0;
  content=0xffffffff;
  return;
}

RTableRec::~RTableRec()
{
  return;
}















