#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <stdio.h>
#include "api.H"
#include "tools.H"
#include "shell.H"
#include "vrevent.H"
#include "lob.H"
#include "forwarder.H"
#include "ffilter.H"
  
#define REMOTE_EXP

#define LOAD_HELP "\
 load help:\n \
 \n \
 load <lob> [arg1] [arg2] [arg3] [...] \n"

#define IFCONFIG_HELP "\
 ifconfig help:\n \
 \n \
 ifconfig\n \
 ifconfig add <ifname> <ip-address> [netmask] [broadcast] \n \
 ifconfig <ifname> delete \n \
 ifconfig <ifname> \n \
 ifconfig <ifname> reset \n \
 ifconfig <ifname> id <newid> \n \
 ifconfig <ifname> ip <newip> \n \
 ifconfig <ifname> nm <newnetmask> \n \
 ifconfig <ifname> bw <bandwidth(Mbps)> [buckettime(bytes)] \n \
 ifconfig <ifname> address <ip> [netmask] [broadcast]\n \
 ifconfig <ifname> connect ...\n \
 ifconfig <ifname> ttx ...\n \
 ifconfig <ifname> trx ...\n \
 ifconfig <ifname> qs \n \
 ifconfig <ifname> qs conns \n \
 ifconfig <ifname> qs list \n \
 ifconfig <ifname> qs create [trio|dsm|droptail|tbf|scheduler|classifier] \n \
 ifconfig <ifname> disconnect \n \
 ifconfig <ifname> connect sol <#no> \n \
 ifconfig <ifname> connect fifo <#no> <{l|r}> \n \
 ifconfig <ifname> connect tunnel <hostname> <dest_port> <src_port> \n \
 ifconfig <ifname> connect ipip <hostname> [mtu] \n \
 ifconfig <ifname> qs chain <component id1> <component id2>\n \
 ifconfig <ifname> qs <component id> reset\n \
 ifconfig <ifname> qs avail\n \
 ifconfig <ifname> qs <component id> status\n \
 ifconfig <ifname> qs <tbf-id> bw <bandwidth> [bucketsize]\n \
 ifconfig <ifname> qs <droptail-id> ql <queuelen>\n \
 ifconfig <ifname> qs <scheduler-id> wt <component id> <weight>\n \
 ifconfig <ifname> qs <trio-id> ql len0 len1 len2 \n \
 ifconfig <ifname> qs <trio-id> mode {hard|linear} \n \
 ifconfig <ifname> qs <dsmarker-id> mark ... as ... \n \
 ifconfig <ifname> qs <dsmarker-id> erase ... as ... \n \
 ifconfig <ifname> qs <scheduler-id> mode {wfq|prr|rr|pwfq} \n"

#define ROUTE_HELP "\
 route help:\n \
 \n \
 route\n \
 route add <ip/nmb> <interface> \n \
 route add <ip> <nm> <interface> \n"

#define SYS_HELP "\
 sys help: \n \
 \n \
 sys events \n \
 sys lobs \n \
 sys filters \n"

#define HELP "\
 for hard coded commands try:\n \
 \n \
 ifconfig help\n \
 route help\n \
 exit (only telnet consoles) \n \
 shutdown \n \
 sys help \n"

Shell::Shell(VR *v):API(v,0)
{
  port=vr->shell_port;		//  we get the parameter from the central VR class
  binary_mode=vr->shell_binary;
  vr=v;
  vr->cfh=1;			// output to stdout
  idstring="shell";
  pw_prompt=0;
  if(vr->shell_password && !binary_mode)pw_prompt=1;
  
  /*
   * process startupscript
   */

  if(vr->shell_startupscript){
    FILE *fp;
    fp=fopen(vr->shell_startupscript,"r");
    if(fp){
      puts("--- processing startup script");

      int argc;
      char *argv[256];
      char line_in_buffer[1024];
      char output_buffer[0xffff];
      memset(line_in_buffer,0,1024);
      memset(output_buffer,0,0xffff);
      
      while(fgets(line_in_buffer,1000,fp)){
	if(line_in_buffer[0]!='#'){
	  strip_trailing(line_in_buffer,"\n \t\r");
	  if(!line_in_buffer[0])continue;
	  argc=split(line_in_buffer,argv);
	  printf("%s> ",line_in_buffer); 
	  interprete(argc, argv, output_buffer, 0xffff);
	  printf("%s> ",output_buffer); 
	}
	memset(line_in_buffer,0,1024);
	memset(output_buffer,0,0xffff);
      }
      puts("--- startup script processed");

    } else {
      printf("Startupfile %s not found\n",vr->shell_startupscript);
    }
  }
  
  if(port)			// Connection auf TCP port
    {  
      fhprintf(vr->cfh,"listening to port %d\n",port);
      
      if(vr->shell_hostname){
	strcpy(hostname,vr->shell_hostname);
      } else {
	gethostname(hostname,256);
      }
      
      memset(&addr,0,sizeof(addr));
      hent=gethostbyname(hostname);
      memcpy(&addr.sin_addr, hent->h_addr, hent->h_length);
      addr.sin_port=htons(port);
      addr.sin_family=AF_INET;
      
      sock=socket(AF_INET, SOCK_STREAM,0);
      if(sock<0) { 
	perror("error at 'socket' call");
	exit(0);
      }
      struct linger ling;
      ling.l_onoff=1;
      ling.l_linger=0;
      
      setsockopt(sock, SOL_SOCKET, SO_LINGER, &ling, sizeof(ling));
      
      if(bind(sock, (struct sockaddr*) &addr, sizeof(addr))!=0) {
	perror("error at 'bind' call");
	exit(0);
      }
      
      if(listen(sock,1)<0){
	perror("error at 'listen' call");
	exit(0);
      }
      
      event=new VRevent(this);
      event->type=IO_READ;	
      event->suspended=0;
      event->suspend_time=HUGE_VAL;
      event->thrown=0;
      event->fh_r=sock;		// wait for new socket connection
      event->fh_w=0 ;		// not used

      vr->vrsched->notify(event);
      sfh=0;			// no open connection

      return;
    } // end port


  // connected to stdin/stdout
  // 
  //
  event = new VRevent(this);
  event->type=IO_READ;	
  event->suspended=0;
  event->suspend_time=HUGE_VAL;
  event->thrown=0;
  event->fh_r=0;		// stdin
  event->fh_w=0 ;		// not used
  vr->vrsched->notify(event);
  banner();
  fflush(stdout);
  return;
}

Shell::~Shell()
{
  if(port){
    if(sfh)TCPclose();
    if(sock)close(sock);
  }
  else {
    vr->vrsched->release(event->event_id);
  }
  
}

int Shell::TCPopen()
{
  if((sfh=accept(sock, (struct sockaddr*) &raddr,(socklen_t*)&raddr_len))<0){
    perror("error at 'accept' call");
    sfh=0;
    return 0;
  }
  return sfh;
}

int Shell::TCPpoll()
{
  int retval;
  
  fd_set rfds;
  struct timeval tv;
  FD_ZERO(&rfds);
  FD_SET(sock, &rfds);
  tv.tv_sec = 0;
  tv.tv_usec = 0;
  
  retval = select(sock+1, &rfds, NULL, NULL, &tv);
  
  return retval;
}


int Shell::TCPclose()
{
  if(sfh){
    close(sfh);
    sfh=0;
    return 0;
  }
  return 1;
}

void Shell::banner()
{
  if(!pw_prompt){
    fhputs("\nMicroVar Shell\n",vr->cfh);
    fhprintf(vr->cfh,"> "); 
    return;
  }
  fhputs("\nMicroVar Shell\n",vr->cfh);
  fhprintf(vr->cfh,"enter password:"); 
  return;
}

int Shell::password_check(char *s)
{
  if(!strcmp(s,vr->shell_password)){
    fhprintf(vr->cfh,"access enabled\n>");
    pw_prompt=0;
    return 0;
  }
  
  fhprintf(vr->cfh,"access denied\nenter password:");
  pw_prompt++;
  return 1;
}

/*
 * this function is called by the fordwarder, if a filter was
 * set up and no alternative callback defined 
 */

int Shell::forward(FFilter *f, Packet *p)
{
  fhprintf(vr->cfh,"Filtered(%u): %s\n", f->cb_handle, pkt2char(p));
  return 0;
}


int Shell::send_RB(VRRB *r)
{
  /* types:
     0 synchronous RBs 
     1 filter packet
  */
  switch(r->rb_type)
    {
    case 0:
      {
	fhprintf(vr->cfh,"Shell Error: got synchronous Result Block id:%u len:%u\n",
		 r->handle, r->len);
	return 0;
      }
    default:
      {
	fhprintf(vr->cfh,"Shell Error: got unknown asynchronous Result Block id:%u len:%u\n",
		 r->handle, r->len);
	return 0;
      }
    }
  
  return 0;
}


int Shell::interrupt(VRevent *e)
{
  if(e->forced){
    e->forced=0;
    puts("Ouch");
  }
  if(e->thrown&IO_READ){
    if(!port) {			// normal shell, no TCP
      char *argv[256];
      char input_buffer[1024];
      char output_buffer[0xffff];
      int len, argc;

      memset(input_buffer,0,1024);
      memset(output_buffer,0,0xffff);
   
      len=read(0, input_buffer, 1024); 
       
      input_buffer[len-1]='\0';

      strip_trailing(input_buffer,"\n \t\r");

      if(pw_prompt){
	password_check(input_buffer);
	return 0;
      }

      if(!input_buffer[0]){
	fhprintf(vr->cfh,">");
	return 0;
      }

      argc=split(input_buffer,argv);
      interprete(argc, argv, output_buffer, 0xffff);
      fhprintf(vr->cfh,"%s> ",output_buffer); 
      input_buffer[0]=0;
      output_buffer[0]=0;
      return 0;
    }
    
    if(!sfh){			// open TCP connection
      sfh=TCPopen();

      fhprintf(vr->cfh,"shell: new TCP connection opened (%d)\n",sfh);
      
      event->fh_r=sfh;		// now listen on open sfh

      vr->cfh=sfh;
      
      if(!binary_mode){
	banner();
      }
      
      return 0;
    }
    
    if(sfh){			// work on TCP connection
      if(!binary_mode){
	char *argv[256];
	int len, argc;
	char input_buffer[1024];
	char output_buffer[0xffff];
	len=read(sfh, input_buffer, 1024); input_buffer[len-1]='\0';
	strip_trailing(input_buffer,"\n \t\r");
	argc=split(input_buffer,argv);

	if(pw_prompt){
	  password_check(input_buffer);
	  if(pw_prompt>vr->shell_password_tries){
	    fhprintf(sfh,"closing connection\n");
	    vr->cfh=1;		// out is stdout again
	    fhprintf(vr->cfh,"shell: TCP connection closed\n");
	    TCPclose();		// close the open socket
	    event->fh_r=sock;	// listen for new connections
	    pw_prompt=0;
	    if(vr->shell_password)pw_prompt=1;
	    return 0;
	  }
	  return 0;
	}
	
	if(argc && !strcmp(argv[0],"exit")){ // somebody closes the connection
	  fhprintf(sfh,"bye\n");
	  vr->cfh=1;		// out is stdout again
	  fhprintf(vr->cfh,"shell: TCP connection closed\n");
	  TCPclose();		// close the open socket
	  event->fh_r=sock;	// listen for new connections
	  pw_prompt=0;
	  if(vr->shell_password)pw_prompt=1;
	  return 0;
	}
	
	interprete(argc, argv, output_buffer,0xffff);
	
	fhprintf(sfh,"%d %s> ",strlen(output_buffer), output_buffer); 
	input_buffer[0]=0;
	output_buffer[0]=0;
      } else {
	//
	//  raw shell
	// 
	int len;
	char input_buffer[1024];
	len=read(sfh, input_buffer, 1024);
	VRRB *rb=process_VRCB((VRCB*)input_buffer);
	write(sfh,rb,rb->len);
	free(rb);
      }
      return 0;
    }    
  }
  return 0;
}


int Shell::split(char *s, char **argv)
{
  int argc=0;
  int m=0;
  
  for(char *p=s;*p;++p){
    if(is_delimiter(*p)){m=0;*p=0; continue;}
    if(m==0){argv[argc++]=p; m=1;}
  }
  return argc;
}


int Shell::is_delimiter(char c)
{
  if(c>='a' && c<='z')return 0;
  if(c>='A' && c<='Z')return 0;
  if(c>='0' && c<='9')return 0;
  if(c=='.' || c==',' || c==':' || c=='?' || c=='/' || 
     c=='#' || c=='-' || c=='|' || c=='&') 
    return 0;
  return 1;
}



// this function interpretes a command line
// finally it only forwards the request to more specilized funtions
//
void Shell::interprete(int argc, char **argv, char *s, int l)
{
  
#ifdef DEBUG_API
  fhprintf(vr->cfh,"\n");
  for(int i=0;i<argc;++i){
    fhprintf(vr->cfh,"DEBUG_API: %d: %s\n",i, argv[i]);
  }
#endif
  
  if(!strcmp(argv[0],"help")){
    strcpy(s,HELP);
    return;
  }

  if(!strcmp(argv[0],"shutdown")){
    exit(0);
    return;
  }
  
  if(!strcmp(argv[0],"ifconfig")){
    return interprete_ifconfig(argc-1, &argv[1], s, l);
  }
  
  if(!strcmp(argv[0],"route")){
    return interprete_route(argc-1, &argv[1],s,l);
  }
  if(!strcmp(argv[0],"sys")){
    return interprete_sys(argc-1, &argv[1],s,l);
  }
  
  if(!strcmp(argv[0],"load") || ! strcmp(argv[0],"!")){
    return interprete_load(argc-1, &argv[1],s,l);
  }
  
  if(!strcmp(argv[0],"test-it")){
    return interprete_test(argc-1, &argv[1],s,l);
  }

  if(!strcmp(argv[0],"attach")){
    return interprete_attach(argc-1, &argv[1],s,l);
  }

  strcpy(s,"Syntax Error\n");
  return;
}


void Shell::interprete_attach(int argc, char **argv, char *s, int l)
{
  int fh_r, fh_w;

  if(argc!=2){
    fhprintf(vr->cfh,"Syntax: attach <fifo_out> <fifo_in> <name>\n");
    return;
  }
  
  fh_w=open(argv[0],O_RDWR|O_NONBLOCK);
  fh_r=open(argv[1],O_RDWR|O_NONBLOCK);
  
  if(fh_w<0){
    fhprintf(vr->cfh,"Error: cannot open %s\n", argv[0]);
    return;
  }

  if(fh_r<0){
    fhprintf(vr->cfh,"Error: cannot open %s\n", argv[1]);
    close(fh_w); 
    return;
  }
  

  new API(vr, fh_r, fh_w);
  
  fhprintf(vr->cfh,"new API attached. <- %s , -> %s\n", argv[0], argv[1]);
      
  return;
}


/*
 * some testroutine
 */

void Shell::interprete_test(int argc, char **argv, char *s, int l)
{

#ifdef DEBUG_API
  fhprintf(vr->cfh,"DEBUG_API: entering testroutine\n");
#endif

  struct FILTER_ADDDEL_CB filter_add_cb;

  memset(&filter_add_cb,0,sizeof(FILTER_ADDDEL_CB));
  filter_add_cb.filter.source_ip=0;
  filter_add_cb.filter.source_nm=0;
  filter_add_cb.filter.dest_ip=0;
  filter_add_cb.filter.dest_nm=0;
  
  filter_add_cb.handle=get_handle();
  filter_add_cb.len=sizeof(struct FILTER_ADDDEL_CB);
  filter_add_cb.id=30;
  filter_add_cb.filter_command=2;
  filter_add_cb.filter.tos=0;
  filter_add_cb.filter.prio=10;
  filter_add_cb.filter.mode=0; 
  filter_add_cb.filter.forwarder=NULL;
  strcpy(filter_add_cb.filter.name,"test");
  filter_add_cb.filter.proto=0;
  //  filter_add_cb.filter.ipopt_type=IPOPT_RA;
  //  filter_add_cb.filter.ipopt_value=0;

#ifdef DEBUG_API
  fhprintf(vr->cfh,"DEBUG_API: starting processing\n");
#endif

  FILTER_RB *rb=(FILTER_RB*)process_VRCB((VRCB*)&filter_add_cb);

#ifdef DEBUG_API
  fhprintf(vr->cfh,"DEBUG_API: processing finished, preparing output\n");
#endif
  
  strncpy(s,filterrec2string(&((FILTER_RB*)rb)->filter),l);

#ifdef DEBUG_API
  fhprintf(vr->cfh,"DEBUG_API: testroutine done\n");
#endif

  return;
}


/*
  * loading og objects
  */

void Shell::interprete_load(int argc, char **argv, char *s, int l)
{
  struct dummy : public VRCB
  {
    unsigned short lob_command;	
    unsigned short flags;	
    char name[32];
    char command_line[1024];		
  } cb;
  
  struct dummz : public VRRB
  {
    struct LOB_INFO_BLOCK m;
  } *rb;
  
  cb.handle=get_handle();
  cb.lob_command=1;
  cb.id=40;
  memset(cb.command_line,0,1024); 
  memset(cb.name,0,32);
  strcpy(cb.name,argv[0]);

  char *p=cb.command_line;
  for(int i=1;i<argc;++i){
    //    printf("arg %d: %s\n",i, argv[i]);
    strcpy(p,argv[i]);p+=strlen(argv[i]); p++;
  }
  cb.len=(unsigned long)p-(unsigned long)&cb;
   
  (VRRB*)rb=process_VRCB((VRCB*)&cb);
   
  sprintf(s,"object '%s' loaded, id is %u, mode is %s\n",
	  rb->m.name,
	  rb->m.lob_id,
	  lobmode2string(rb->m.mode)
	  );
  return;
}

/////////////////////////////////////////////////////////////////////////
//
// interprete the route command
//
void Shell::interprete_route(int argc, char **argv, char *s, int l)
{

  // no arguments list all routes

  if(argc==0)
    {			// route
      ROUTE_VRCB cb;
      cb.handle=get_handle();
      cb.len=sizeof(ROUTE_VRCB);
      cb.id=20;
      cb.command=3;
      
      VRRB *rb=process_VRCB((VRCB*)&cb);
      ROUTE_RECORD *rr=(ROUTE_RECORD*)(rb+1);
      int n=(rb->len-sizeof(VRRB))/sizeof(ROUTE_RECORD);
      int w=0;
      strcpy(s,"no routes\n");
      for(int i=0;i<n;++i)
	{
	  IF_VRCB idcb; 
	  IF_INFORMATION_RB *idrb;
	  idcb.handle=get_handle(); 
	  idcb.len=sizeof(IF_VRCB);
	  idcb.id=10; 
	  idcb.cmd=1; 
	  idcb.ifno=rr[i].ifid;
	  idrb=(IF_INFORMATION_RB*)process_VRCB((VRCB*)&idcb);

	  w+=sprintf(s+w,"route ");

	  if(rr[i].dest_nm==0 && rr[i].dest_ip==0)
	    {
	      w+=sprintf(s+w,"default to ");
	    }
	  else 
	    {
	      w+=sprintf(s+w,"%s/",n2char(rr[i].dest_ip));
	      w+=sprintf(s+w,"%s to ",n2char(rr[i].dest_nm));
	    }
	 
	  w+=sprintf(s+w,"%s(%d) \n",idrb->name,rr[i].ifid);
	  free(idrb);
	  
	}
      free(rb);
      return;
    }
  
  // help argument, copy help string to output buffer

  if(argc==1 && !strcmp(argv[0],"help"))
    {
      strcpy(s,ROUTE_HELP);
      return;
    }
  
  // add command add new routes
  
  
  if(argc>=3 && !strcmp(argv[0],"add"))
    {
      ROUTE_RECORD_VRCB cb;
      ROUTE_RECORD_VRRB *rb;
      IF_INFORMATION_RB *nametorb=NULL;

      // initialize the route add control block 

      cb.source_ip=cb.dest_ip=0xffffffff;
      cb.proto=cb.tos=0xffff;
      cb.handle=get_handle();
      cb.len=sizeof(ROUTE_RECORD_VRCB);
      cb.command=1;
      cb.id=20;
      
      /* 
	 // some debugging
	 for(int zi=0;zi<argc;++zi)
	 {
	 printf("%d: %s\n", zi, argv[zi]);
	 }
      */

      // four argument: route add 10.1.1.1 255.255.255.0 if0

      if(argc==4)
	{		
	  if(!is_ifname(argv[3]))
	    {
	      strcpy(s,"no such interface\n");
#ifdef REMOTE_EXP
	      strcpy(s,"-2 no such interface\n");
#endif	    
	      return;
	    }

	  nametorb=get_interface_by_name(argv[3]);

	  if(!is_ipadd(argv[1]) || !is_ipadd(argv[2]))
	    {
	      strcpy(s,"Syntax Error\n");
#ifdef REMOTE_EXP
	      strcpy(s,"-3 Syntax Error\n");
#endif	    
	      return;
	    }

	  cb.dest_ip=char2n(argv[1]);
	  cb.dest_nm=char2n(argv[2]);

	}
      
      // three argument: route add 10.1.1.0/24 if0
      
      if(argc==3)
	{	
	  if(!is_ifname(argv[2]))
	    {
	      strcpy(s,"no such interface\n");
#ifdef REMOTE_EXP
	      strcpy(s,"-2 no such interface\n");
#endif
	      return;
	    }

	  nametorb=get_interface_by_name(argv[2]);
	  if(!strcmp(argv[1],"default"))
	    {
	      char2ipnm("0.0.0.0/0",&cb.dest_ip, &cb.dest_nm);
	    }
	  else
	    {
	      if(char2ipnm(argv[1],&cb.dest_ip, &cb.dest_nm))
		{
		  strcpy(s,"Syntax Error\n");
#ifdef REMOTE_EXP
		  strcpy(s,"-3 Syntax Error\n");
#endif	    
		  return;
		}
	    }
	}

      if(nametorb==NULL)
	{
	  strcpy(s,"no such interface\n");
#ifdef REMOTE_EXP
	  strcpy(s,"-2 no such interface\n");
#endif
	  return;
	}

      // add the interface number to the control block

      cb.ifid=nametorb->ifno;

      // we have all information extracted from the command line
      // and can now send the route add reqauest over the API
      
      rb=(ROUTE_RECORD_VRRB*)process_VRCB(&cb);

      // create dom output
      
      if(rb->ifid==0xffff)
	{
	  strcpy(s,"error adding route\n");
#ifdef REMOTE_EXP 
	  strcpy(s,"-1 error adding route\n");
#endif
	}
      else
	{
	  int w=0;
	  
#ifdef REMOTE_EXP
	  w+=sprintf(s+w,"0 ");
#endif
	  w+=sprintf(s+w,"route ");
	  
	  if(rb->dest_ip==0 && rb->dest_nm==0)
	    {
	      w+=sprintf(s+w,"default ");
	    }
	  else 
	    {
	      w+=sprintf(s+w,"%s/",n2char(rb->dest_ip));
	      w+=sprintf(s+w,"%s ",n2char(rb->dest_nm));
	    }
	  w+=sprintf(s+w,"to %s added\n",nametorb->name);
	  
	}

      
      free(nametorb);
      free(rb);
      return;
    }
  
  
  
  if(argc>1 && !strcmp(argv[0],"del")){
    ROUTE_RECORD_VRCB cb;
    ROUTE_RECORD_VRRB *rb;
    IF_INFORMATION_RB *nametorb=NULL;
    
    cb.source_ip=cb.dest_ip=0xffffffff;
    cb.proto=cb.tos=0xffff;
    cb.handle=get_handle();
    cb.len=sizeof(ROUTE_RECORD_VRCB);
    cb.command=2;		// delete route code
    cb.id=20;

    // four argument: route del 10.1.1.1 255.255.255.0 if0
     
    if(argc==4)
      {		
	nametorb=get_interface_by_name(argv[3]);
	cb.dest_ip=char2n(argv[1]);
	cb.dest_nm=char2n(argv[2]);
      }
     
    // three argument: route del 10.1.1.0/24 if0
     
    if(argc==3)
      {	
	nametorb=get_interface_by_name(argv[2]);
	if(!strcmp(argv[1],"default"))
	  {
	    char2ipnm("0.0.0.0/0",&cb.dest_ip, &cb.dest_nm);
	  }
	else
	  {
	    char2ipnm(argv[1],&cb.dest_ip, &cb.dest_nm);
	  }
      }
     
    if(nametorb==NULL)
      {
	strcpy(s,"no such interface\n");
#ifdef REMOTE_EXP
	strcpy(s,"-2 no such interface\n");
#endif
	return;
      }
     
    // add the interface number to the control block
     
    cb.ifid=nametorb->ifno;
     
    // we have all information extracted from the command line
    // and can now send the route delete reqauest over the API
   
    rb=(ROUTE_RECORD_VRRB*)process_VRCB(&cb);

    if(rb->ifid==0xffff)
      {
	strcpy(s,"error deleting route\n");
#ifdef REMOTE_EXP
	strcpy(s,"-1 error deleting route\n");
#endif
      }
    else 
      {
	int w=0;
#ifdef REMOTE_EXP
	w+=sprintf(s+w,"route %s/",n2char(rb->dest_ip));
#else
	w+=sprintf(s+w,"0 route %s/",n2char(rb->dest_ip));
#endif
	w+=sprintf(s+w,"%s to ",n2char(rb->dest_nm));
	w+=sprintf(s+w,"%s deleted\n",nametorb->name);
      }
    
    free(nametorb);
    free(rb);
     
    return;
  }

  strcpy(s,"syntax error\n");
  return;
}



void Shell::interprete_sys(int argc, char **argv, char *s, int l)
{
  if(argc==0){		
    strcpy(s,SYS_HELP);
    return;
  }

  if(argc==1 && !strcmp(argv[0],"events")){		
    VRCB cb;
    cb.handle=get_handle();
    cb.len=sizeof(VRCB);
    cb.id=21;			// list events

    VRRB *rb=process_VRCB(&cb);
    struct EVENT_REC *ev = (struct EVENT_REC*)(rb+1);

    int l=sprintf(s,"Registered Events:\n");
    for(int i=0;i*sizeof(struct EVENT_REC)+sizeof(VRCB)<rb->len;++i){
      l+=EVENT_REC2string(&(ev[i]),s+l);
    }
    free(rb);
    return;
  }

  if(argc==1 && !strcmp(argv[0],"lobs")){		

    struct dummy0 : public VRCB {
      unsigned short locmd;
    } cb;

    struct dummy1 : public VRRB {
      LOB_INFO_BLOCK ibl;
    } *rb;

    cb.locmd=10;
    cb.handle=get_handle();
    cb.len=sizeof(struct dummy0);
    cb.id=40;			// loadable object related

    (VRRB*)rb=process_VRCB(&cb);
    
    if(rb->len==sizeof(VRRB)){
      strcpy(s,"no objects loaded\n");
      return;
    }

    LOB_INFO_BLOCK *b=&(rb->ibl);
    int n=(rb->len-sizeof(VRRB))/sizeof(LOB_INFO_BLOCK);
    int m=0;
    for(int i=0; i<n; ++i){
      m+=sprintf(s+m,"name %s, id %u, mode %s, size %lu kb, up %2.2f min\n",
		 b[i].name,
		 b[i].lob_id,
		 lobmode2string(b[i].mode),
		 b[i].size/1024,
		 b[i].uptime/60
		 );
    }
    free(rb);
  }

  if(argc==1 && !strcmp(argv[0],"filters")){		
    
    struct FILTER_CB cb;
    cb.handle=get_handle();
    cb.len=sizeof(FILTER_CB);
    cb.id=30;			// filter specific
    cb.filter_command=0;	// list filters
    
    struct dummy0 : public VRRB{
      struct FILTER_REC filter;
    } *rb;
    
    (VRRB*)rb=process_VRCB((VRCB*)&cb);

    if(rb->len==sizeof(VRRB)){
      strcpy(s,"no filter installed\n");
      free(rb);
      return;
    }

    int n=(rb->len-sizeof(VRRB))/sizeof(FILTER_REC);
    struct FILTER_REC *f = &rb->filter;
    
    int k=0;
    k+=sprintf(s+k,"Filters:\n");
    for(int i=0;i<n;++i){
      k+=sprintf(s+k,"%s, ", f[i].name);
      k+=sprintf(s+k,"source %s, ", ipnm2char(f[i].source_ip,f[i].source_nm));
      k+=sprintf(s+k,"dest %s\n", ipnm2char(f[i].dest_ip,f[i].dest_nm));
      k+=sprintf(s+k,"proto %s\n", proto_id2char(f[i].proto));
    }
    free(rb);
    return;
  }
  
  return;
}

char * Shell::lobmode2string(unsigned short mode)
{
  static char b[256];
  *b=0;
  if(!mode){strcpy(b,"-"); return b;}
  if(mode&LOB_MODE_RUN)strcat(b,"RUN ");
  if(mode&LOB_MODE_STAY)strcat(b,"STAY ");
  if(mode&LOB_MODE_THREAD)strcat(b,"THREAD ");
  return b;
}


///////////////////////////////////////////////////////////////////////
// 
// interprete the ifconfig command and return any result in
// s up the the length l 
//
//
void Shell::interprete_ifconfig(int argc, char **argv, char *s, int l)
{
  memset(s,0,l); strcpy(s,"\n");

  // no argument give list all interfaces
 
  if(argc<=0){			// ifconfig
    int l=0;
    VRCB cb;
    VRRB *rb;
    cb.handle=get_handle();
    cb.len=sizeof(VRCB);
    cb.id=3;
    rb=process_VRCB(&cb);	// get interface inidizes

    if(rb->len==sizeof(VRRB)){

#ifdef REMOTE_EXP	
	  strcpy(s,"-2 no interfaces defined\n");
#else
	  strcpy(s,"no interfaces defined\n");
#endif

      free(rb);
      return;
    }
    
    IF_VRCB icb;
    IF_INFORMATION_RB *irb;
      
    for(int i=sizeof(VRRB);i<rb->len;++i){ // process single interface
      icb.handle=get_handle();
      icb.len=sizeof(IF_VRCB);
      icb.id=10;
      icb.cmd=1;
      icb.ifno=((char*)rb)[i];

      irb=(IF_INFORMATION_RB*)process_VRCB((VRCB*)&icb);

      if(irb->ifno==-1){
	strcpy(s+l,"error\n"); 
	l+=strlen("error\n");
      } else {
	strcpy(s+l,"---\n"); 
	l+=strlen("---\n");
	l=IF_INFORMATION_RB2string(irb,s,l);
      }
      free(irb);
    }

    free(rb);
    return;
  }

  // help argument copy the help string to the output buffer

  if(argc==1 && !strcmp(argv[0],"help")){	// ifconfig help 
    strncpy(s,IFCONFIG_HELP,l-1);
    return;
  }

  // add command adds a new interface

  if(argc>1 && !strcmp(argv[0],"add")) // add interface
    {	
      IF_ADDINTERFACE_CB cb;
      IF_INFORMATION_RB *rb;


      // check for interface name
      
      if(is_ifname(argv[1])==0)
	{
#ifdef REMOTE_EXP	
	  strcpy(s,"-3 Syntax Error\n");
#else
	  strcpy(s,"Syntax Error\n");
#endif
	  return;
	} 

      // Check whether parameters are ip addresses

      if( (argc==5 && (!is_ipadd(argv[2]) || !is_ipadd(argv[3]) || !is_ipadd(argv[4]))) ||
	  (argc==4 && (!is_ipadd(argv[2]) || !is_ipadd(argv[3]))) ||
	  (argc==3 && (!is_ipadd(argv[2]))))
	{
#ifdef REMOTE_EXP	
	  strcpy(s,"-3 Syntax Error\n");
#else
	  strcpy(s,"Syntax Error\n");
#endif
	  return;
	}
	  
      switch(argc)
	{
	case 5:			// ifconfig add if0 10.1.1.1 255.255.255.0 10.1.1.255
	  strcpy(cb.name,argv[1]);
	  cb.ip=char2n(argv[2]);
	  cb.nm=char2n(argv[3]);
	  cb.bc=char2n(argv[4]);
	  break;
	case 4:			// ifconfig add if0 10.1.1.1 255.255.255.0 
	  strcpy(cb.name,argv[1]);
	  cb.ip=char2n(argv[2]);
	  cb.nm=char2n(argv[3]);
	  cb.bc=char2n("000.000.000.255")|char2n(argv[2]);
	  break;
	case 3:			// ifconfig add if0 10.1.1.1
	  strcpy(cb.name,argv[1]);
	  cb.ip=char2n(argv[2]);
	  cb.nm=char2n("255.255.255.0");
	  cb.bc=char2n("000.000.000.255")|char2n(argv[2]);
	  break;
	default:
#ifdef REMOTE_EXP	
	  strcpy(s,"-3 Syntax Error\n");
#else
	  strncpy(s,IFCONFIG_HELP,l-1);
#endif
	  return;
	}
      
      // okay all parameters parsed and assumed okay, now lets
      // check whether that name exists already
      // try to get interface control block for interface with that
      // number to check whether interface exists already
      
      IF_INFORMATION_RB *test_rb=get_interface_by_name(argv[1]);
      if(test_rb!=NULL){		// block returned -> iterface exists
	free(test_rb);

#ifdef REMOTE_EXP	
	strcpy(s,"-2 Interface already exists\n");
#else
	strcpy(s,"Error: Interface already exists\n");
#endif

	return;
      }
      
      // okay all done, complete the control block and 
      // sent it over the API
      
      cb.handle=get_handle();
      cb.id=1;
      cb.len=sizeof(IF_ADDINTERFACE_CB);
      
      rb=(IF_INFORMATION_RB*)process_VRCB((VRCB*)&cb);
      
      // display the result block
      
      s[0]=0;
      IF_INFORMATION_RB2string(rb,s,0);
      free(rb);
      
      return;
    }

  
  // okay, all keywords are parsed, now we assume the first keyword
  // is an interface id and we watn to trigger some interface related 
  // commands
  
  IF_INFORMATION_RB ifinforb;	// will contain the spe info block
  IF_INFORMATION_RB *ifrb;	// for temporary things
  
  // check whether argv[0] is interface id
  
  ifrb=get_interface_by_name(argv[0]);
  if(ifrb==NULL){
#ifdef REMOTE_EXP	
	strcpy(s,"-2 no such interface\n");
#else
	strcpy(s,"no such interface\n");
#endif
    return;
  }
  memcpy(&ifinforb,ifrb,sizeof(IF_INFORMATION_RB));
  free(ifrb);

  // argv[1] is interface and ifinfoorb points to the approriate information 
  // block.
  // interface specific commands

  if(argc==1){			// [ifconfig] if0
    IF_INFORMATION_RB2string(&ifinforb,s);
    return;
  }

  // deleting Interface

  if(argc==2 && !strcmp(argv[1],"delete")){	// [ifconfig] if0 delete
    IF_DELETEINTERFACE_CB ifcb;
    IF_DELETEINTERFACE_RB *ifdrb;
    ifcb.handle=get_handle();
    ifcb.id=2;			// delete interface
    ifcb.len=sizeof(IF_DELETEINTERFACE_CB);
    ifcb.ifno=ifinforb.ifno;
    
    ifdrb=(IF_DELETEINTERFACE_RB*)process_VRCB((VRCB*)&ifcb);
    
    if(ifdrb->ifno==0xffff)
      {
#ifdef REMOTE_EXP	
	strcpy(s,"-2 no such interface\n");
#else
	strcpy(s,"no such interface\n");
#endif
      } 
    else 
      {
#ifdef REMOTE_EXP	
	strcpy(s,"0 interface deleted\n");
#else
	strcpy(s,"interface deleted\n");
#endif
      }
    free(ifdrb);		// delete new one
    return;
  }
  // resetting counters

  if(argc==2 && !strcmp(argv[1],"reset")){	// [ifconfig] if0 reset
    IF_VRCB ifcb;
    ifcb.handle=get_handle();
    ifcb.id=10;			// interface specific
    ifcb.ifno=ifinforb.ifno;
    ifcb.len=sizeof(IF_VRCB);
    ifcb.cmd=2;			// reset

    ifrb=(IF_INFORMATION_RB*)process_VRCB((VRCB*)&ifcb);
    s[0]='\n';
    free(ifrb);			// delete new one
    return;
  }

  // changing interface name

  if(argc==3 && !strcmp(argv[1],"id")){	// [ifconfig] if0 id newid
    IF_VRCB *ifcb = (IF_VRCB*)malloc(sizeof(IF_VRCB)+256);
    ifcb->handle=get_handle();
    ifcb->id=10;		// interface specific
    ifcb->ifno=ifinforb.ifno;
    ifcb->len=sizeof(IF_VRCB)+10;
    ifcb->cmd=101;		// edit name
    strcpy((char*)(ifcb+1),argv[2]);
    
    ifrb=(IF_INFORMATION_RB*)process_VRCB((VRCB*)&ifcb);
    s[0]='\n';
    free(ifrb);			// delete new one
    free(ifcb);			// delete control block
    return;
  }

  // changing ip address

  if(argc==3 && !strcmp(argv[1],"ip")){	// [ifconfig] if0 ip newid 
    struct dummy0 : public IF_VRCB{
      unsigned long ip;
      unsigned long nm;
      unsigned long bc;
    };
    struct dummy0 ifcb;
    ifcb.handle=get_handle();
    ifcb.id=10;		// interface specific
    ifcb.ifno=ifinforb.ifno;
    ifcb.len=sizeof(struct dummy0);
    ifcb.cmd=102;		// edit ip address
    ifcb.ip=0xffffffff;
    ifcb.nm=0xffffffff;
    ifcb.bc=0xffffffff;
    ifcb.ip=char2n(argv[2]);
    ifrb=(IF_INFORMATION_RB*)process_VRCB((VRCB*)&ifcb);
    if(ifrb==NULL){
      strcpy(s, "PANIC:  error\n");
      return;
    }
    IF_INFORMATION_RB2string(ifrb,s);
    free(ifrb);			// delete new one
    return;
  }

  // changing nm address

  if(argc==3 && !strcmp(argv[1],"nm")){	// [ifconfig] if0 ip newid 
    struct dummy0 : public IF_VRCB{
      unsigned long ip;
      unsigned long nm;
      unsigned long bc;
    };
    struct dummy0 ifcb;
    ifcb.handle=get_handle();
    ifcb.id=10;		// interface specific
    ifcb.ifno=ifinforb.ifno;
    ifcb.len=sizeof(struct dummy0);
    ifcb.cmd=102;		// edit ip address
    ifcb.ip=0xffffffff;
    ifcb.nm=0xffffffff;
    ifcb.bc=0xffffffff;
    ifcb.nm=char2n(argv[2]);
    ifrb=(IF_INFORMATION_RB*)process_VRCB((VRCB*)&ifcb);
    if(ifrb==NULL){
      strcpy(s, "PANIC:  error\n");
      return;
    }
    IF_INFORMATION_RB2string(ifrb,s);
    free(ifrb);			// delete new one
    return;
  }


  // changing interface bandwidth

  if((argc==3 || argc==4) && !strcmp(argv[1],"bw")){	// [ifconfig] if0 bw 10000
    struct dummy0 : public IF_VRCB{
      unsigned long bw;
      unsigned long bs;
    } ifcb;

    ifcb.handle=get_handle();
    ifcb.id=10;		// interface specific
    ifcb.ifno=ifinforb.ifno;
    ifcb.len=sizeof(struct dummy0);
    ifcb.cmd=105;		// set bandwisth/bucketsize
    ifcb.bw=(unsigned long)(atof(argv[2])*1000000.0/8.0); // in bytes per second
    ifcb.bs=0xffffffff;
    if(argc==4){
      ifcb.bs=(unsigned long)atol(argv[3]);
    }
    ifrb=(IF_INFORMATION_RB*)process_VRCB((VRCB*)&ifcb);
    sprintf(s,"Interface bandwidth %f Mbps, bucketsize %lu bytes\n", ifrb->bandwidth*8.0/1000000.0, ifrb->bucketsize);
    free(ifrb);		
    return;
  }


  // Disconnececting interface

  if((argc == 2) && !strcmp(argv[1],"disconnect")){
    struct dummy : public IF_VRCB
    {
      char ctype;			// type of connection
      unsigned long cval[3];
    } cb;
    cb.id=10;
    cb.cmd=109;
    cb.ifno=ifinforb.ifno;
    cb.handle=get_handle();
    cb.len=sizeof(cb);
    cb.ctype=0;			// disconnect it;
    cb.cval[0]=0;		// unused 
    cb.cval[1]=0;	        // unused
    cb.cval[2]=0;		// unused 
    
    ifrb=(IF_INFORMATION_RB*)process_VRCB((VRCB*)&cb);
    IF_INFORMATION_RB2string(ifrb, s);
    free(ifrb);
    return;
  }
  
  
  // connecting the interface to softlink device
  
  if((argc > 3) && !strcmp(argv[1],"connect")){
    
    if((argc == 4) && !strcmp(argv[2],"sol")){ // [ifconfig] if0 connect sol 1 
      struct dummy : public IF_VRCB
      {
	char ctype;			// type of connection
	unsigned long cval[3];
      } cb;
      cb.id=10;
      cb.cmd=109;
      cb.ifno=ifinforb.ifno;
      cb.handle=get_handle();
      cb.len=sizeof(cb);
      cb.ctype=1;			//- connection type is IPC
      cb.cval[0]=0;			//- unused 
      cb.cval[1]=atoi(argv[3]);	        //- fifo number
      cb.cval[2]=0;
      
      ifrb=(IF_INFORMATION_RB*)process_VRCB((VRCB*)&cb);
      IF_INFORMATION_RB2string(ifrb, s);
      free(ifrb);
      return;
    }
    
    if((argc == 5) && !strcmp(argv[2],"fifo")){	// [ifconfig] if0 connect fifo 1 r
      struct dummy : public IF_VRCB
      {
	char ctype;			// type of connection
	unsigned long cval[3];		
      } cb;
      cb.handle=get_handle();
      cb.id=10;
      cb.ifno=ifinforb.ifno;
      cb.cmd=109;
      cb.len=sizeof(cb);
      cb.ctype=2;			//- connection type is IPC
      cb.cval[0]=0;			//- unused 
      cb.cval[1]=atoi(argv[3]);	//- fifo number
      cb.cval[2]=42;
      if(argv[4][0]=='l' || argv[4][0]=='0') 
	cb.cval[2]=1;
      if(argv[4][0]=='r' || argv[4][0]=='1')
	cb.cval[2]=2;

      if(cb.cval[2]==42){
	strcpy(s,"missing or wrong end descriptor\n");
	return;
      } 

      ifrb=(IF_INFORMATION_RB*)process_VRCB((VRCB*)&cb);
      IF_INFORMATION_RB2string(ifrb, s);
      free(ifrb);
      return;
    }

    if((argc == 6) && !strcmp(argv[2],"tunnel")){	// [ifconfig] if0 connect tunnel host 8000 8001

      struct dummy : public IF_VRCB
      {
	char ctype;			// type of connection
	unsigned long cval[3];
      } cb;

      cb.handle=get_handle();
      cb.len=sizeof(cb);
      cb.id=10;
      cb.cmd=109;
      cb.ifno=ifinforb.ifno;
      cb.ctype=3;			//- connection type is tunnel
      puts(argv[3]);
      cb.cval[0]=hostname2n(argv[3]);	//- host ip

      cb.cval[1]=atoi(argv[4]);	//- tx port
      cb.cval[2]=atoi(argv[5]);	//- rx port

      if(cb.cval[1]==0 || cb.cval[2]==0){
	strcpy(s,"missing or wrong port number(s)\n");
	return;
      }

      ifrb=(IF_INFORMATION_RB*)process_VRCB((VRCB*)&cb);
      IF_INFORMATION_RB2string(ifrb, s);
      free(ifrb);
      return;
    }

    if((argc == 4 || argc == 5) && !strcmp(argv[2],"ipip")){	// [ifconfig] if0 connect ipip [mtu]
      
      struct dummy : public IF_VRCB
      {
	char ctype;			// type of connection
	unsigned long cval[3];
      } cb;

      cb.handle=get_handle();
      cb.len=sizeof(cb);
      cb.id=10;
      cb.cmd=109;
      cb.ifno=ifinforb.ifno;
      cb.ctype=4;			//- connection type is ipip
      puts(argv[3]);
      cb.cval[0]=hostname2n(argv[3]);	//- host ip
      
      cb.cval[1]=0; cb.cval[2]=0;
      if(argc==5)cb.cval[2]=atoi(argv[4]);	//- mtu

      ifrb=(IF_INFORMATION_RB*)process_VRCB((VRCB*)&cb);
      IF_INFORMATION_RB2string(ifrb, s);
      free(ifrb);
      return;
    }


    strcpy(s,"wrong connection specifier\n");
    return;
  }

  //
  // all the queueing system stuff
  //

  if(argc>=3 && (!strcmp(argv[1],"qs") || !strcmp(argv[1],"qc"))){	
    
    
    if(argc==3 && !strcmp(argv[2],"avail")){
      int k=0;
      k+=sprintf(s+k,"Available Component Types:\n");
      k+=sprintf(s+k,"  droptail\n");
      k+=sprintf(s+k,"  tbf\n");
      k+=sprintf(s+k,"  classifier\n");
      k+=sprintf(s+k,"  scheduler\n");
      k+=sprintf(s+k,"  dsm\n");
      k+=sprintf(s+k,"  trio\n");
      return;
    }

    if(argc==3 && !strcmp(argv[2],"list")){
      IF_Q_VRCB qcb;
      qcb.id=10;		// interface specific
      qcb.cmd=120;		// the queueing system
      qcb.handle=get_handle();
      qcb.len=sizeof(IF_Q_VRCB);
      qcb.ifno=ifinforb.ifno;
      qcb.qcmd=1;		// list all components
      
      VRRB *qrb=process_VRCB((VRCB*)&qcb);

      int k=0;
      k+=sprintf(s+k,"Queueing Components:\n");
      for(unsigned int i=0;i<(qrb->len-sizeof(VRRB))/2;i+=2){
	k+=sprintf(s+k,"%d %s\n",((unsigned short*)(qrb+1))[i],qid2string(((unsigned short*)(qrb+1))[i+1]));
      }
      free(qrb);
      return;
    }


    if(argc==3 && !strcmp(argv[2],"conns")){
      IF_Q_VRCB qcb;
      qcb.id=10;		// interface specific
      qcb.cmd=120;		// the queueing system
      qcb.handle=get_handle();
      qcb.len=sizeof(IF_Q_VRCB);
      qcb.ifno=ifinforb.ifno;
      qcb.qcmd=2;		// list all connections

      VRRB *qrb=process_VRCB((VRCB*)&qcb);

      struct dummy0 {
	unsigned short si, st, di, dt;
      } *cs;

      cs=(struct dummy0*)(qrb+1);
      int k=0;
      k+=sprintf(s+k,"Connected Components:\n");
      for(unsigned int i=0;i<(qrb->len-sizeof(VRRB))/sizeof(struct dummy0);i++){
	k+=sprintf(s+k,"  %d/%s -> %d/%s\n",
		   cs[i].si, qid2string(cs[i].st), 
		   cs[i].di, qid2string(cs[i].dt));
      }
      
      free(qrb);
      return;
    }


    if(argc==4 && !strcmp(argv[2],"create")){
      struct dummy : public IF_Q_VRCB {
	unsigned short v;
      } qcb; 
      qcb.id=10;		// interface specific
      qcb.cmd=120;		// the queueing system
      qcb.handle=get_handle();
      qcb.len=sizeof(IF_Q_VRCB)+sizeof(unsigned short);
      qcb.ifno=ifinforb.ifno;
      qcb.qcmd=3;		// create component
      qcb.v=0xffff;
      
      if(!strcmp(argv[3],"droptail")){ // type value 1
	qcb.v=1;
      }
      if(!strcmp(argv[3],"tbf")){ // type value 1
	qcb.v=2;
      }
      if(!strcmp(argv[3],"classifier")){ // type value 1
	qcb.v=3;
      }
      if(!strcmp(argv[3],"scheduler")){ // type value 1
	qcb.v=4;
      }
      if(!strcmp(argv[3],"trio")){ // type value 1
	qcb.v=5;
      }
      if(!strcmp(argv[3],"dsm")){ // type value 1
	qcb.v=6;
      }

      if(qcb.v==0xffff){
	strcpy(s,"no such component type\n");
	return;
      }

      VRRB *qrb=process_VRCB((VRCB*)&qcb);

      if(*(unsigned short*)(qrb+1)==0xffff){
	sprintf(s,"error creating component %s\n",argv[3]);
	free(qrb);
	return;
      }
      sprintf(s,"Component created with id %u\n",*(unsigned short*)(qrb+1));
      free(qrb);
      return;
    }

    if(argc==5 && !strcmp(argv[2],"chain")){
      struct dummy : public IF_Q_VRCB {
	unsigned short v[2];
      } qcb; 

      struct dummy2 : public VRRB {
	unsigned short v[2];
      }; 
      struct dummy2 *qrb;

      qcb.id=10;		// interface specific
      qcb.cmd=120;		// the queueing system
      qcb.handle=get_handle();
      qcb.len=sizeof(qcb);
      qcb.ifno=ifinforb.ifno;
      qcb.qcmd=5;		// create connection
      qcb.v[0]=atoi(argv[3]);
      qcb.v[1]=atoi(argv[4]);

      qrb=(struct dummy2*)process_VRCB((VRCB*)&qcb);

      if(qrb->v[0]==0xffff){
	sprintf(s,"error connecting components (%d)\n",qrb->v[1]);
	return;
      }
      free(qrb);
      sprintf(s,"components connected\n");
      return;    
    }

    /*
     * okay all commands directly accessing one queueing component follow
     * here
     *
     * base syntax is: ifconfig if0 qs #id command
     * e.g.:           ifconfig if0 qs #id status
     * e.g.:           ifconfig if0 qs #id reset
     * e.g.:           ifconfig if0 qs #if do_something
     *
     */

    // the third arg is a number
      

    if(argc>=4 && (atoi(argv[2]) || argv[2][0]=='0')){
      int component_id=atoi(argv[2]);
      int interface_no=ifinforb.ifno;

      /*
       * status query
       */

#ifdef DEBUG_API
      fhprintf(vr->cfh,"DEBUG_API: queueing system command\n");
#endif

      if(!strcmp(argv[3],"status")){
#ifdef DEBUG_API
	fhprintf(vr->cfh,"DEBUG_API: got status command\n");
#endif

	IF_QS_VRCB qcb;
	qcb.id=10;		// interface specific
	qcb.cmd=120;		// the queueing system
	qcb.handle=get_handle();
	qcb.len=sizeof(IF_QS_VRCB);
	qcb.ifno=interface_no;
	qcb.qcmd=10;		// component specific
	qcb.component_id=component_id;
	qcb.component_command=2; // query status 

	QS_STATISTICS_VRRB *rb=(QS_STATISTICS_VRRB*)process_VRCB((VRCB*)&qcb);
	if(rb->len==sizeof(VRRB)){
	  strcpy(s, "Error: command not supported by component\n");
	  return;
	}
	
	switch(rb->component_type)
	  {
	  case 0:		// root component
	    {
	      QS_ROOT_STATISTICS_VRRB *b=(QS_ROOT_STATISTICS_VRRB*)rb;
	      int l=0;
	      l+=sprintf(s+l,"Root Component:\n");
	      l+=sprintf(s+l,"enques     %lu    deques     %lu\n",
			 b->enques, b->deques);
	      l+=sprintf(s+l,"stored     %lu    dropped    %lu\n",
			 b->global_stored,b->global_dropped);
	      free(rb);
	      return;
	    }
	  case 1:		// droptail component
	    {
	      QS_DROPTAIL_STATISTICS_VRRB *b=(QS_DROPTAIL_STATISTICS_VRRB*)rb;
	      int l=0;
	      l+=sprintf(s+l,"Droptail Queue:\n");
	      l+=sprintf(s+l,"dropped    %lu    deques requests     %lu\n",
			 b->dropped_packets, b->deque_requests);
	      l+=sprintf(s+l,"deque hits %lu    deque fails         %lu\n",
			 b->deque_hits, b->deque_fails);
	      l+=sprintf(s+l,"queuelen   %u    \n",b->queuelen);
	      free(rb);
	      return;
	    }
	  case 5:		// TRIO component
	    {
	      QS_TRIO_STATISTICS_VRRB *b=(QS_TRIO_STATISTICS_VRRB*)rb;
	      int l=0;
	      l+=sprintf(s+l,"Trio Queue:\n");
	      l+=sprintf(s+l,"queuelen:  %8u %8u %8u\n",
			 b->queuelen[0], b->queuelen[1], b->queuelen[2]);
	      l+=sprintf(s+l,"dropped:   %8lu %8lu %8lu\n",
			 b->dropped_packets[0],b->dropped_packets[1], b->dropped_packets[2]);
	      l+=sprintf(s+l,"pkts:      %u\n",
			 b->pkts);
	      l+=sprintf(s+l,"deque hits %lu    deque fails         %lu\n",
			 b->deque_hits, b->deque_fails);
	      switch(b->mode){
	      case 0:
		l+=sprintf(s+l,"mode is 'hard'\n"); break;
	      case 1:
		l+=sprintf(s+l,"mode is 'linear'\n"); break;
	      default:
		l+=sprintf(s+l,"mode is unknown\n"); break;
	      }
	      free(rb);
	      return;
	    }
	  case 2:		// Token bucket filter component
	    {
	      QS_TBF_STATISTICS_VRRB *b=(QS_TBF_STATISTICS_VRRB*)rb;
	      int l=0;
	      l+=sprintf(s+l,"Token Bucket Filter:\n");
	      l+=sprintf(s+l,"enque hits        %lu    enque fails     %lu\n",
			 b->enque_hits, b->enque_fails);
	      l+=sprintf(s+l,"enque empty       %lu    \n", 
			 b->enque_empty);
	      l+=sprintf(s+l,"deque hits        %lu    deque fails     %lu\n",
			 b->deque_hits, b->deque_fails);
	      l+=sprintf(s+l,"deque empty       %lu    \n", 
			 b->deque_empty);
	      l+=sprintf(s+l,"bucket overflows  %lu    \n", 
			 b->bucket_overflows);
	      l+=sprintf(s+l,"bandwidth         %5.4f Mbps\n", 
			 b->bandwidth/1000000.0);
	      l+=sprintf(s+l,"bucketsize        %5.4f Mbit\n", 
			 b->bucketsize/1000000.0);
	      free(rb);
	      return;
	    }
	  case 4:		// Scheduler
	    {
	      QS_SCHEDULER_STATISTICS_VRRB *r=(QS_SCHEDULER_STATISTICS_VRRB*)rb;
      
	      GEN_SCHEDULER_RECORD *rec=(GEN_SCHEDULER_RECORD *)(r+1); // appended to VRRB 
	      
	      int l=0;
	      switch(r->mode){
	      case 0: 
		l+=sprintf(s+l,"Scheduler Mode: Priority Round Robin\n");
		break;
	      case 1:
		l+=sprintf(s+l,"Scheduler Mode: Weighted Fair Queueing\n");
		break;
	      case 2:
		l+=sprintf(s+l,"Scheduler Mode: Round Robin\n");
		break;
	      default:
		l+=sprintf(s+l,"Scheduler Mode: -unknown-\n");
	      }
	      
	      int n=(rb->len-sizeof(QS_SCHEDULER_STATISTICS_VRRB))/sizeof(GEN_SCHEDULER_RECORD);
	      
	      if(n)l+=sprintf(s+l,"preceding components:\n");
	      for(int i=0;i<n; ++i){
		l+=sprintf(s+l,"  %d/%s w/p: %u deq: %lu\n",
			   rec[i].id, qid2string(rec[i].type), rec[i].misc, rec[i].deque_hits);
	      }

	      if(!n){
		l+=sprintf(s+l,"sorry, no preceding components\n");
	      }

	      free(rb);
	      return;
	    }
	  case 3:		// Classifier
	    {
	      QS_CLASSIFIER_STATISTICS_VRRB *r=(QS_CLASSIFIER_STATISTICS_VRRB*)rb;
      	      GEN_CLASSIFIER_RECORD *rec=(GEN_CLASSIFIER_RECORD *)(r+1); // appended to VRRB 
	      int l=0;
	      l+=sprintf(s+l,"Classifier:\n");

	      int n=(rb->len-sizeof(QS_CLASSIFIER_STATISTICS_VRRB))/sizeof(GEN_CLASSIFIER_RECORD);

	      printf("%d %d %d\n", rb->len, sizeof(QS_CLASSIFIER_STATISTICS_VRRB),n);
	      
	      for(int i=0;i<n; ++i){
		l+=sprintf(s+l,"%d %s %u %u\n",
			   rec[i].id, qid2string(rec[i].type), 0, 0);
	      }
	      
	      if(!n){
		l+=sprintf(s+l,"sorry, no classifier rules applied\n");
	      }
	      
	      free(rb);
	      return;
	    }

	  case 6:		// DS marker
	    {
#ifdef DEBUG_API
	      fhprintf(vr->cfh,"DEBUG_API: adding dsm rule\n");
#endif

	      QS_DSMARKER_STATISTICS_VRRB *r=(QS_DSMARKER_STATISTICS_VRRB*)rb;
      	      DS_MARKER_RECORD *rec=(DS_MARKER_RECORD *)(r+1); // appended to VRRB 
	      int l=0;
	      l+=sprintf(s+l,"Differentiated Services Marker (dsm):\n");
	      
	      int n=(rb->len-sizeof(QS_DSMARKER_STATISTICS_VRRB))/sizeof(DS_MARKER_RECORD);
	      
	      //	      printf("%d %d %d\n", rb->len, sizeof(QS_DSMARKER_STATISTICS_VRRB),n);
	      
	      for(int i=0;i<n; ++i){
		l+=sprintf(s+l,"---\n");
		l+=sprintf(s+l,"source: %s/", n2char(rec[i].source_ip));
		l+=sprintf(s+l,"%s     ", n2char(rec[i].source_nm));
		l+=sprintf(s+l,"dest: %s/", n2char(rec[i].dest_ip));
		l+=sprintf(s+l,"%s\n", n2char(rec[i].dest_nm));
		l+=sprintf(s+l,"proto: %u     tos: %u\n", rec[i].proto, rec[i].tos);
		if(rec[i].service==100){
		  l+=sprintf(s+l,"service: Expedited Forwarding\n"); 
		  continue;
		}
		
		if(rec[i].service>200 && rec[i].service<205){
		  l+=sprintf(s+l,"service:  Assured Forwarding Class %d with\n", rec[i].service-200);
		  l+=sprintf(s+l,"\t %f Mbps and a bucketsize of %lu bytes for low drop precedence\n",
			     rec[i].param[0]*8.0/1000000.0, rec[i].param[1]);
		  l+=sprintf(s+l,"\t %f Mbps and a bucketsize of %lu bytes for medium drop precedence\n",
			     rec[i].param[2]*8.0/1000000.0, rec[i].param[3]);
		  continue;
		}

		l+=sprintf(s+l,"service: UNKNOWN (internal id: %d)\n",rec[i].service);
	      }
	      
	      if(!n){
		l+=sprintf(s+l,"sorry, no marker rules applied\n");
	      }
	      
	      free(rb);
	      return;
	    }
	    
	  default:
	    strcpy(s,"unknown or unimplemented component, cannot interprete statistical data\n");
	    free(rb);
	    return;
	  }
      } // end of status

      /*
       * resetting a components counters
       */

      if(!strcmp(argv[3],"reset")){
	IF_QS_VRCB qcb;
	qcb.id=10;		// interface specific
	qcb.cmd=120;		// the queueing system
	qcb.handle=get_handle();
	qcb.len=sizeof(IF_QS_VRCB);
	qcb.ifno=interface_no;
	qcb.qcmd=10;		// component specific
	qcb.component_id=component_id;
	qcb.component_command=1; // reset counters

	VRRB *rb=process_VRCB((VRCB*)&qcb);
	free(rb);
      } // end of reset

      /*
       * now follow the commands different for single components
       * so we get the component type for syntax checks
       */

#ifdef DEBUG_API
      fhprintf(vr->cfh,"DEBUG_API: get control block for component type\n");
#endif

      IF_QS_VRCB comp_scb;
      comp_scb.id=10;		// interface specific
      comp_scb.cmd=120;		// the queueing system
      comp_scb.handle=get_handle();
      comp_scb.len=sizeof(IF_QS_VRCB);
      comp_scb.ifno=interface_no;
      comp_scb.qcmd=10;		// component specific
      comp_scb.component_id=component_id;
      comp_scb.component_command=2; // query status 
      
      QS_STATISTICS_VRRB *comp_srb=(QS_STATISTICS_VRRB*)process_VRCB((VRCB*)&comp_scb);
      if(comp_srb->len==sizeof(VRRB)){
	sprintf(s, "Error: command not supported by component\n");
	free(comp_srb);
	return;
      }

      /*
       * now we can process the component dependent commands
       */
      
      
      // TOKEN BUCKET FILTER  bandwidth and bucketsize setting
      //
      // ifconfif if0 3 bw 2.0 2048
      if(!strcmp(argv[3],"bw") && comp_srb->component_type==2){
	tbf_configure(argc-4,argv+4,interface_no, component_id,s);
	free(comp_srb);
	return;
      }

      
      // DROPTAIL  queuelen setting
      //
      // [ifconfig] if0 qs 1 ql 64
      //
      if(!strcmp(argv[3],"ql") && comp_srb->component_type==1){
	droptail_configure(argc-4,argv+4,interface_no, component_id,s);
	free(comp_srb);
	return;
      }

      // SCHEDULER mode setting
      //
      // [ifconfig] if0 qs 1 mode [prr|wfq|rr|pwfq]
      // [ifconfig] if0 qs 1 wt 12 200

      if(!strcmp(argv[3],"mode") && comp_srb->component_type==4){
	scheduler_mode_setting(argc-4,argv+4,interface_no, component_id,s);
	free(comp_srb);
	return;
      }
      if(!strcmp(argv[3],"wt") && comp_srb->component_type==4){
	scheduler_wt_setting(argc-4,argv+4,interface_no, component_id,s);
	free(comp_srb);
	return;
      }
      
      // DSMARKER commands
      //
      // expected command
      // ifconfig if0 qs 3 mark  source 0.0.0.0/0 as af{1,2,3,4}[:b1:t1:b2:t2]|ef
      // ifconfig if0 qs 3 erase source 0.0.0.0/0 as af{1,2,3,4}[:b1:t1:b2:t2]|ef

      if(!strcmp(argv[3],"mark") && comp_srb->component_type==6){
	dsmarker_rule_set(argc-4,argv+4, interface_no, component_id,s);
	free(comp_srb);
	return;
      }
      if(!strcmp(argv[3],"erase") && comp_srb->component_type==6){
	dsmarker_rule_erase(argc-3,argv+4, interface_no, component_id,s);
	free(comp_srb);
	return;
      }

      // TRIO queulen  bandwidth and bucketsize setting
      //
      // ifconfig if0 3 ql 64 32 16 
      if(!strcmp(argv[3],"ql") && comp_srb->component_type==5){
	dsmarker_rule_set(argc-4,argv+4,interface_no, component_id,s);
	free(comp_srb);
	return;
      }

      free(comp_srb);
     
    } // end of component configuration (argv[2] is a number)
    
    
    
  } // end of queueing stuff
  
  // change addresses

  if(argc>=3 && !strcmp(argv[1],"address")){	
    strcpy(s, "not implemented yet\n");
    return;
  }

  // connect the interface

  if(argc>=3 && !strcmp(argv[1],"connect")){	
    strcpy(s, "not implemented yet\n");

    return;
  }

  // iptrans

  if(argc>=3 && !strcmp(argv[1],"trx")){	
    strcpy(s, "not implemented yet\n");
    return;
  }

  // iptrans

  if(argc>=3 && !strcmp(argv[1],"ttx")){	
    strcpy(s, "not implemented yet\n");
    return;
  }

  strcpy(s,IFCONFIG_HELP);
  return;
}


int Shell::IF_INFORMATION_RB2string(IF_INFORMATION_RB *r, char *s, int l)
{
  char c;

#ifdef REMOTE_EXP
  l+=sprintf(s+l,"0 %s ",r->name);
  l+=sprintf(s+l,"%s ",n2char(r->ip));
  l+=sprintf(s+l,"%s ",n2char(r->netmask));
  l+=sprintf(s+l,"%d\n",r->ctype);
  return l;
#endif

  //
  l+=sprintf(s+l,"name:           %14s\n",r->name);
  l+=sprintf(s+l,"ip-address:     %14s\n",n2char(r->ip));
  l+=sprintf(s+l,"netmask:        %14s\n",n2char(r->netmask));
  l+=sprintf(s+l,"broadcast:      %14s                          \n",n2char(r->broadcast));
  l+=sprintf(s+l,"bandwidth(bps): %14lu   bucketsize(bytes): %14lu\n",r->bandwidth*8,r->bucketsize);
  l+=sprintf(s+l,"drops:          %14lu   errors:            %14lu\n",r->drops, r->errors);
  l+=sprintf(s+l,"rx:             %14lu   tx:                %14lu\n",r->rx, r->tx);

  if(r->iptrans_rx_nm!=0 || r->iptrans_tx_nm!=0){
    l+=sprintf(s+l,"rx-t-ip         %14s   ",n2char(r->iptrans_rx_ip));
    l+=sprintf(s+l,"rx-t-nm            %14s\n",n2char(r->iptrans_rx_nm));
    l+=sprintf(s+l,"rx-t-val        %14s   ",n2char(r->iptrans_rx_value));
    l+=sprintf(s+l,"rx-t-pat           %14s\n",n2char(r->iptrans_rx_pattern));
    l+=sprintf(s+l,"tx-t-ip         %14s   ",n2char(r->iptrans_tx_ip));
    l+=sprintf(s+l,"tx-t-nm            %14s\n",n2char(r->iptrans_tx_nm));
    l+=sprintf(s+l,"tx-t-pat        %14s\n",n2char(r->iptrans_tx_pattern));
  }
  // connectors

  switch(r->ctype){
  case 0: 
    l+=sprintf(s+l,"connection      %14s\n","none");
    break;
  case 1:
    l+=sprintf(s+l,"connection      %14s   ","softlink");
    l+=sprintf(s+l,"id:                %14lu\n",r->cval[1]);
    break;
  case 2:
    l+=sprintf(s+l,"connection      %14s   ","local");
    l+=sprintf(s+l,"id:                %14lu\n",r->cval[2]);
    c='l'; if(r->cval[2]==2)c='r';
    l+=sprintf(s+l,"end of cable                 %c\n",c);
    break;
  case 3:
    l+=sprintf(s+l,"connection      %14s   ","tunnel");
    l+=sprintf(s+l,"host:              %14s\n",n2char(r->cval[0]));
    l+=sprintf(s+l,"tx_port         %14lu   ",r->cval[1]);
    l+=sprintf(s+l,"rx_port            %14lu\n",r->cval[2]);
    break;
  case 4:
    l+=sprintf(s+l,"connection      %14s   ","ipiptunnel");
    l+=sprintf(s+l,"destination:    %14s\n",n2char(r->cval[0]));
    l+=sprintf(s+l,"mtu             %14lu\n",r->cval[2]);
    break;

  }
  return l;
}


char* Shell::qid2string(int n)
{
  switch(n){
  case 0: return "root";
  case 1: return "droptail";
  case 2: return "tbf";
  case 3: return "classifier";
  case 4: return "scheduler";
  case 5: return "trio";
  case 6: return "dsm";
  }
  return "unkown";
}



int Shell::EVENT_REC2string(struct EVENT_REC *ev, char *s, int l)
{
  char type[8];
  char thrown[8];

  *type=*thrown=0;
  
  if(ev->type&IO_READ)strcat(type,"R");
  if(ev->type&IO_WRITE)strcat(type,"W");
  if(ev->type&TIMER)strcat(type,"T");
  if(ev->type&SINGLE)strcat(type,"S");

  if(ev->thrown&IO_READ)strcat(thrown,"R");
  if(ev->thrown&IO_WRITE)strcat(thrown,"W");
  if(ev->thrown&TIMER)strcat(thrown,"T");
  if(ev->thrown&SINGLE)strcat(thrown,"S");
  
  l+=sprintf(s+l,"  %6s susp %.2f type %3s thrown %3s read %u write %u hits %lu\n",
	     ev->comment, ev->suspend_time, type, thrown, ev->fh_r, ev->fh_w, ev->hits);

  return l;
}






int Shell::parse_dsmarkerrule(int argc, char **argv, DS_MARKER_RECORD *dsr)
{
  int flag=0;
  memset(dsr,0,sizeof(DS_MARKER_RECORD));
  dsr->tos=0xffff;
  dsr->proto=0xffff;
  dsr->service=0xffff;
  
  for(int i=0;i<argc;++i){
    if(!strcmp(argv[i],"source") && i+1<argc){
      ++i;
      flag++;
      char2ipnm(argv[i], &dsr->source_ip, &dsr->source_nm);
      continue;
    }
    if(!strcmp(argv[i],"dest") && i+1<argc){
      ++i;
      flag++;
      char2ipnm(argv[i], &dsr->dest_ip, &dsr->dest_nm);
      continue;
    }
    if(!strcmp(argv[i],"proto") && i+1<argc){
      ++i;
      flag++;
      dsr->proto=atoi(argv[i]);
      continue;
    }
    if(!strcmp(argv[i],"tos") && i+1<argc){
      ++i;
      flag++;
      dsr->tos=atoi(argv[i]);
      continue;
    }
    if(!strcmp(argv[i],"as") && i+1<argc){
      ++i;
      if(!strncmp(argv[i],"ef",2)){
	dsr->service=100;
	break;
      }

      if(!strncmp(argv[i],"af",2)){
	if(argv[i][2]<'1' || argv[i][2]>'4')break;
	dsr->service=argv[i][2]-'1'+201;
	int z=0;
	for(int j=strlen(argv[i]);j;j--)if(argv[i][j]==':')z++;
	char *u=strdup(argv[i]);

	switch(z)
	  {
	  case 2:{
	    char *a, *b;
	    a=strchr(u,':');++a;
	    b=strchr(a,':');*b=0;++b;
	    dsr->param[0]=(unsigned long)(atof(a)*1000000.0/8.0);
	    dsr->param[2]=(unsigned long)(atof(b)*1000000.0/8.0);
	    break;
	  }
	  case 4:{
	    char *a, *b, *c, *d;
	    a=strchr(u,':');++a;
	    b=strchr(a,':');*b=0;++b;
	    c=strchr(b,':');*c=0;++c;
	    d=strchr(c,':');*d=0;++d;
	    dsr->param[0]=(unsigned long)(atof(a)*1000000.0/8.0);
	    dsr->param[1]=atol(b);
	    dsr->param[2]=(unsigned long)(atof(c)*1000000.0/8.0);
	    dsr->param[3]=atol(d);
	    break;
	  }
	  } // switch
	free(u);
	break;
      }
    }
  } // for
  if(!flag)return 1;				   // flag true when spec set
  return 0;
}






int Shell::dsmarker_rule_set(int argc, char **argv, int if_no, int comp_id, char *s)
{

  /*
    for(int i=0;i<argc;++i){
    printf("%d %s\n",i,argv[i]);
    }
  */

  if(argc<4){
    strcpy(s,"Error: too few arguments\n");
    return 1;					   // error
  }
  
  struct dummy0 : public IF_QS_VRCB 
  {
    DS_MARKER_RECORD ds;
  } qcb;
  
  qcb.id=10;			// interface specific
  qcb.cmd=120;			// the queueing system
  qcb.handle=get_handle();
  qcb.len=sizeof(struct dummy0);
  qcb.ifno=if_no;
  qcb.qcmd=10;			// component specific
  qcb.component_id=comp_id;	// the id of the component
  qcb.component_command=611;	// add DS marker rule
  
  if(parse_dsmarkerrule(argc, argv, &qcb.ds) || qcb.ds.service==0xffff){
    sprintf(s,"Syntax Error\n");
    return 1;
  }
  
  // request senden und antwort auswerten
  
  struct dummy1 : public QS_DSMARKER_STATISTICS_VRRB
  {
    DS_MARKER_RECORD rec;
  };
  
  struct dummy1 *rb=(struct dummy1*)process_VRCB((VRCB*)&qcb);

  int l=0;
  l+=sprintf(s+l,"---\n");
  l+=sprintf(s+l,"source: %s/", n2char(rb->rec.source_ip));
  l+=sprintf(s+l,"%s     ", n2char(rb->rec.source_nm));
  l+=sprintf(s+l,"dest: %s/", n2char(rb->rec.dest_ip));
  l+=sprintf(s+l,"%s\n", n2char(rb->rec.dest_nm));
  l+=sprintf(s+l,"proto: %u     tos: %u\n", rb->rec.proto, rb->rec.tos);
  
  if(rb->rec.service==100){
    l+=sprintf(s+l,"service: Expedited Forwarding\n"); 
    free(rb);
    return 0;
  }
  
  if(rb->rec.service>200 && rb->rec.service<205){
    l+=sprintf(s+l,"service:  Assured Forwarding Class %d with\n", rb->rec.service-200);
    l+=sprintf(s+l,"\t %f Mbps and a bucketsize of %lu bytes for low drop precedence\n",
	       rb->rec.param[0]*8.0/1000000.0, rb->rec.param[1]);
    l+=sprintf(s+l,"\t %f Mbps and a bucketsize of %lu bytes for medium drop precedence\n",
	       rb->rec.param[2]*8.0/1000000.0, rb->rec.param[3]);
    free(rb);
    return 0;
  }
  
  l+=sprintf(s+l,"service: UNKNOWN (internal id: %d)\n",rb->rec.service);
  
  free(rb);
  return 0;
} // end of ds marker rule setup






int Shell::dsmarker_rule_erase(int argc, char **argv, int if_no, int comp_id, char *s)
{
  if(argc<4){strcpy(s,"Error: too few arguments\n"); return 1;}					   // error
  
  struct dummy0 : public IF_QS_VRCB 
  {
    DS_MARKER_RECORD ds;
  } qcb;
  
  qcb.id=10;		// interface specific
  qcb.cmd=120;		// the queueing system
  qcb.handle=get_handle();
  qcb.len=sizeof(struct dummy0);
  qcb.ifno=if_no;
  qcb.qcmd=10;		// component specific
  qcb.component_id=comp_id;
  qcb.component_command=612; // erase DS marker rule
  
  if(parse_dsmarkerrule(argc, argv, &qcb.ds) || qcb.ds.service==0xffff){
    sprintf(s,"Syntax Error\n");
    return 1;
  }
  
  // request senden und antywort auswerten
  
  struct dummy1 : public QS_DSMARKER_STATISTICS_VRRB
  {
    DS_MARKER_RECORD rec;
  };
  
  struct dummy1 *rb=(struct dummy1*)process_VRCB((VRCB*)&qcb);
  
  if(rb->len!=sizeof(struct dummy1)){
    sprintf(s,"no matching rule found\n");
    free(rb);
    return 1;
  }


  int l=0;
  l+=sprintf(s+l,"---\n");
  l+=sprintf(s+l,"source: %s/", n2char(rb->rec.source_ip));
  l+=sprintf(s+l,"%s     ", n2char(rb->rec.source_nm));
  l+=sprintf(s+l,"dest: %s/", n2char(rb->rec.dest_ip));
  l+=sprintf(s+l,"%s\n", n2char(rb->rec.dest_nm));
  l+=sprintf(s+l,"proto: %u     tos: %u\n", rb->rec.proto, rb->rec.tos);
  
  if(rb->rec.service==100){
    l+=sprintf(s+l,"service: Expedited Forwarding\n"); 
    free(rb);
    return 0;
  }
  
  if(rb->rec.service>200 && rb->rec.service<205){
    l+=sprintf(s+l,"service:  Assured Forwarding Class %d with\n", rb->rec.service-200);
    l+=sprintf(s+l,"\t %f Mbps and a bucketsize of %lu bytes for low drop precedence\n",
	       rb->rec.param[0]*8.0/1000000.0, rb->rec.param[1]);
    l+=sprintf(s+l,"\t %f Mbps and a bucketsize of %lu bytes for medium drop precedence\n",
	       rb->rec.param[2]*8.0/1000000.0, rb->rec.param[3]);
    free(rb);
    return 0;
  }
  
  l+=sprintf(s+l,"service: UNKNOWN (internal id: %d)\n",rb->rec.service);
  
  free(rb);
  return 0;
} // end of ds marker rule erase







int Shell::trio_configure(int argc, char **argv, int if_no, int comp_id, char *s)
{
  if(argc<3){strcpy(s,"Error: too few arguments\n");
  return 1;
  }

  struct dummy0 : public IF_QS_VRCB 
  {
    unsigned short mode;
    unsigned short queuelen[3];
  } qcb;
  
	
  qcb.id=10;		// interface specific
  qcb.cmd=120;		// the queueing system
  qcb.handle=get_handle();
  qcb.len=sizeof(IF_QS_VRCB);
  qcb.ifno=if_no;
  qcb.qcmd=10;		// component specific
  qcb.component_id=comp_id;
  qcb.component_command=710; // TRIO CONFIG
  qcb.mode=0;
  qcb.queuelen[0]=atoi(argv[4]);
  qcb.queuelen[1]=atoi(argv[5]);
  qcb.queuelen[2]=atoi(argv[6]);
  
  if(qcb.queuelen[0]>qcb.queuelen[1] || qcb.queuelen[1]>qcb.queuelen[2]){
    sprintf(s,"Queuelen settings invalid\n");
    return 1;
  }

  QS_TRIO_STATISTICS_VRRB *rb=(QS_TRIO_STATISTICS_VRRB*)process_VRCB((VRCB*)&qcb);
  
  sprintf(s,"TRIO queue configuration: %d %d %d\n",
	  rb->queuelen[0], rb->queuelen[1], rb->queuelen[2]);
  
  free(rb);
  return 0;
} // end of TRIO configuration








int Shell::scheduler_wt_setting(int argc, char **argv, int if_no, int comp_id, char *s)
{
  if(argc!=3){strcpy(s,"Error: wrong number of arguments\n"); return 1;	}
  
  struct dummy0 : public IF_QS_VRCB 
  {
    unsigned short icomp_id;
    unsigned short icomp_wt;
  } qcb;
	
  qcb.icomp_id=atoi(argv[0]);
  qcb.icomp_wt=atoi(argv[1]);
  
  qcb.id=10;		// interface specific
  qcb.cmd=120;		// the queueing system
  qcb.handle=get_handle();
  qcb.len=sizeof(IF_QS_VRCB);
  qcb.ifno=if_no;
  qcb.qcmd=10;		// component specific
  qcb.component_id=comp_id;
  qcb.component_command=411; // change scheduler weight  
  QS_SCHEDULER_STATISTICS_VRRB *rb=
    (QS_SCHEDULER_STATISTICS_VRRB*)process_VRCB((VRCB*)&qcb);
  
  /*
   * Error
   */ 
  if(rb->len!=sizeof(QS_SCHEDULER_STATISTICS_VRRB)+
     sizeof(GEN_SCHEDULER_RECORD)){
    sprintf(s,"Error: no such preceding component\n");
    free(rb);
    return 1;
  }
  
  int k=0;
  switch(rb->mode){
  case 0:
    k+=sprintf(s+k,"Scheduler (%u):\nMode PRR\n",rb->component_id);
    break;
  case 1:
    k+=sprintf(s+k,"Scheduler (%u):\nMode WFQ\n",rb->component_id);
    break;
  case 2:
    k+=sprintf(s+k,"Scheduler (%u):\nMode RR\n",rb->component_id);
    break;
  case 3:
    k+=sprintf(s+k,"Scheduler (%u):\nMode PWFQ\n",rb->component_id);
    break;
  }
  
  GEN_SCHEDULER_RECORD *rec=(GEN_SCHEDULER_RECORD*)(rb+1);
  k+=sprintf(s+k,"prec. components weight:\n");
  k+=sprintf(s+k,"  %d/%s w/p: %u deq: %lu\n",
	     rec->id, qid2string(rec->type), 
	     rec->misc, rec->deque_hits);
  free(rb);
  return 1;
} // end of scheduler mode setting



IF_INFORMATION_RB* Shell::get_interface_by_name(char *ifname)
{
  // create the control block

  IF_INFORMATION_RB *ifrb;
  IF_IDNAMEQUERY_CB ib;
  memset(&ib,0,sizeof(IF_IDNAMEQUERY_CB));
  ib.len=sizeof(IF_IDNAMEQUERY_CB);
  ib.handle=get_handle();
  ib.id=4;			// query interface by name
  strncpy(ib.name,ifname,8);
  ifrb=(IF_INFORMATION_RB*)process_VRCB(&ib);
  if(ifrb->ifno==-1){
    free(ifrb); 
    return NULL;		// no such interface
  }
  return ifrb;
}



int Shell::scheduler_mode_setting(int argc, char **argv, int if_no, int comp_id, char *s)
{
  if(argc!=5){strcpy(s,"Error: wrong number of arguments\n"); return 1;}

  struct dummy0 : public IF_QS_VRCB 
  {
    unsigned short mode;
  } qcb;
  
  qcb.mode=0xffff;
  if(!strcmp(argv[4],"prr"))qcb.mode=0;
  if(!strcmp(argv[4],"wfq"))qcb.mode=1;
  if(!strcmp(argv[4],"rr"))qcb.mode=2;
  if(!strcmp(argv[4],"pwfq"))qcb.mode=3;
  
  if(qcb.mode==0xffff){
    strcpy(s,"Error: undefined scheduler mode\n");
    return 1;
  }

  qcb.id=10;		// interface specific
  qcb.cmd=120;		// the queueing system
  qcb.handle=get_handle();
  qcb.len=sizeof(struct dummy0);
  qcb.ifno=if_no;
  qcb.qcmd=10;		// component specific
  qcb.component_id=comp_id;
  qcb.component_command=410; // change scheduler modo  
  QS_SCHEDULER_STATISTICS_VRRB *rb=
    (QS_SCHEDULER_STATISTICS_VRRB*)process_VRCB((VRCB*)&qcb);
  
  switch(rb->mode){
  case 0:
    sprintf(s,"Scheduler (%u):\nMode PRR\n",rb->component_id);
    break;
  case 1:
    sprintf(s,"Scheduler (%u):\nMode WFQ\n",rb->component_id);
    break;
  case 2:
    sprintf(s,"Scheduler (%u):\nMode RR\n",rb->component_id);
    break;
  case 3:
    sprintf(s,"Scheduler (%u):\nMode PWFQ\n",rb->component_id);
    break;
  }
  free(rb);
  return 0;
} // end of scheduler mode setting








int Shell::droptail_configure(int argc, char **argv, int if_no, int comp_id, char *s)
{

  if(argc!=1){
    strcpy(s,"Error: wrong number of arguments\n");
    return 1;
  }

  struct dummy0 : public IF_QS_VRCB 
  {
    unsigned long ql;
  } qcb;
  
  qcb.ql=atoi(argv[0]);
  
  qcb.id=10;		// interface specific
  qcb.cmd=120;		// the queueing system
  qcb.handle=get_handle();
  qcb.len=sizeof(struct dummy0);
  qcb.ifno=if_no;
  qcb.qcmd=10;		// component specific
  qcb.component_id=comp_id;
  qcb.component_command=310; // configure droptail queue  
  QS_DROPTAIL_STATISTICS_VRRB *rb=
    (QS_DROPTAIL_STATISTICS_VRRB*) process_VRCB((VRCB*)&qcb);
  
  sprintf(s,"Droptail Queue (%u):\nQueuelen %u packets \n", 
	  rb->component_id, rb->queuelen);
  free(rb);
  return 0;
} // end of droptail queuelen setting









int Shell::tbf_configure(int argc, char **argv, int if_no, int comp_id, char *s)
{  
  if(argc<2){
    strcpy(s,"Error: too few arguments\n");
    return 1;
  }
  
  struct dummy0 : public IF_QS_VRCB 
  {
    unsigned long bw;
    unsigned long bs;
  } qcb;
  
  qcb.bs=0xffffffff;
  qcb.bw=(unsigned long)(atof(argv[4])*1000000.0); // bits per second
  if(argc==6)
    qcb.bs=(unsigned long)(atof(argv[5])*1000000.0);  // in bits per second
  
  qcb.id=10;		// interface specific
  qcb.cmd=120;		// the queueing system
  qcb.handle=get_handle();
  qcb.len=sizeof(struct dummy0);
  qcb.ifno=if_no;
  qcb.qcmd=10;		// component specific
  qcb.component_id=comp_id;
  qcb.component_command=210;
  QS_TBF_STATISTICS_VRRB *rb=(QS_TBF_STATISTICS_VRRB*)process_VRCB((VRCB*)&qcb);
  
  sprintf(s,"Token Bucket Filter (%u):\nBandwidth: %5.3f Mbps Bucketsize: %5.3f Mbit\n", 
	  rb->component_id, rb->bandwidth/1000000.0, rb->bucketsize/1000000.0); 
  
  free(rb);
  return 0;
} // end of tbf bandwidth and bucketsize set


// returns whehter char * points to a string
// containing somethin which looks like a interface
// name (if0, if1 ...)
//
int Shell::is_ifname(char *s)
{
  if(!s)return 0;
  int l=strlen(s);
  if(l<3 || l>4)return 0;
  if(s[0]!='i' || s[1]!='f')return 0;
  if(atoi(s+2))return 1;
  if(s[2]=='0' && s[3]=='\0')return 1;
  return 0;
}


int Shell::is_ipadd(char *s)
{
  int i,k,n; 
  
  n=strlen(s);
  if(n<7 || n>15)return 0;
  
  for(i=0,k=0;i<n;++i){
    if(s[i]=='.')++k;
  }
  if(k!=3)return 0;
  return 1;
}


char* Shell::filterrec2string(FILTER_REC *f)
{
  static char s[1024];
  int l=0;
  l+=sprintf(s+l,"source:  %s/",n2char(f->source_ip));
  l+=sprintf(s+l,"%s\n",n2char(f->source_nm));
  l+=sprintf(s+l,"dest:    %s/",n2char(f->dest_ip));
  l+=sprintf(s+l,"%s\n",n2char(f->dest_nm));
  l+=sprintf(s+l,"proto:   %d\n",f->proto);
  l+=sprintf(s+l,"tos:     %d\n",f->tos);
  l+=sprintf(s+l,"comment: %s\n",f->name);
  l+=sprintf(s+l,"mode:    ");
  if(!f->mode)l+=sprintf(s+l,"COPY\n"); else l+=sprintf(s+l,"MOVE\n");
  l+=sprintf(s+l,"handle:  %u\n",f->cb_handle);

  return s;
}



