#include <stdio.h>
#include <getopt.h>
#include <stdlib.h>
#include <string.h>
#include <intercon/udp_connection.H>
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>
#include "tools.H"

int print_help()
{
  puts("Softlink Shunt (c) baf");
  puts("");
  puts("Options:");
  puts("     -s device               the softlink devicefile");
  puts("     -d address:port         destination address and port");
  puts("     -p port                 port to receive packets on");
  puts("     -b Mbps                 speed of the shunt in Megabit per second");
  puts("     -l bytes                bucketsize");

  puts("");
  return 0;
}

int main(int argc, char **argv)
{
  char *device=0;
  char *destination=0;
  unsigned short port=0;
  double bw=0;
  unsigned long bs=0, bl=0;
  double last_called;
  int solfh;
  UDP_Sender   *sender;
  UDP_Receiver *receiver;

  while(1){
    int i=getopt(argc,argv,"s:d:p:b:l:");
    if(i<=0)break;
    switch(i)
      {
      case 's':
	device=strdup(optarg);
	break;
      case 'd':
	destination=strdup(optarg);
	break;
      case 'p':
	port=atoi(optarg);
	break;
      case 'b':
	bw=atof(optarg);
	break;
      case 'l':
	bs=atol(optarg);
	break;
      default:
	print_help();
	return 1;
      }
  }


  if(device==0 || destination==0 || port==0 || bw==0 || strchr(destination,':')==0){
    print_help();
    return 1;
  }
  char *c=strchr(destination,':'); *c='\0'; ++c;

  if(atoi(c)==0){
    print_help();
    return 1;
  }

  solfh=open(device,O_RDWR);
  if(solfh<=0){
    print_help();
    return 1;
  }
    
  sender = new UDP_Sender(destination, atoi(c));
  receiver = new UDP_Receiver(port);
  
  int fh_max;
  if(receiver->get_fh()>=solfh)
    fh_max=receiver->get_fh()+1; 
  else
    fh_max=solfh+1;


  // the io loop

  last_called=gettimeasdouble();
  while(1) {
    char buffer[2048];
    fd_set rfds;
    struct timeval tv;
    
    FD_ZERO(&rfds);
    FD_SET(receiver->get_fh(), &rfds);
    FD_SET(solfh, &rfds);
    
    tv.tv_sec=5;
    tv.tv_usec=0;

    bl+=(unsigned long)((gettimeasdouble()-last_called)*bw*125000);
    if(bl>bs)bl=bs;
    
    int r=select(fh_max,&rfds, NULL, NULL, &tv);
    
    if(r){
      if(FD_ISSET(solfh, &rfds)){
	if(bl>2048){
	  int len=read(solfh, buffer, 2048);
	  sender->sendbuff(buffer,len);
	}
      }
      if(FD_ISSET(receiver->get_fh(), &rfds)){
	int len=read(receiver->get_fh(), buffer, 2048);
	write(solfh, buffer, len);
      }
    }
  }
  return 0;
}
