#ifndef _socks_h_
#define _socks_h_

#define MAX_NUMBER_OF_RAW_SOCKETS 32
#define MAX_NUMBER_OF_UDP_SOCKETS 32

/*
 * Flags for the socket states
 */

#define VRSOCK_READ      1	// Socket is in read mode
#define VRSOCK_WRITE     2	// write mode
#define VRSOCK_DISABLED  4	// socket defunct, because of error
#define VRSOCK_PARENT    8	// sock is owned by forwarder
#define VRSOCK_DIRECT    16	// no pipes but hardwired function call

/*
 *  return values for read and write
 */

#define VRSOCK_BADMODE  -1
#define VRSOCK_ERROR    -2
#define VRSOCK_MISSOPT  -3

/*
 * Option Codes  for the setopt function
 */
#define VRSOCK_DEST       1
#define VRSOCK_PROTO      2
#define VRSOCK_HDRINC     3
#define VRSOCK_NONBLOCK   4
#define VRSOCK_SOURCE     5
#define VRSOCK_TTL        6
#define VRSOCK_LOCALPORT  7
#define VRSOCK_REMOTEPORT 8


class VR;
class ProtocolStack;

class Sock
{
public:
  Sock();
  Sock(VR*);
  virtual ~Sock();
  virtual int read(char*,int) = 0;
  virtual int write(char*,int) = 0;
  virtual int setopt(int, unsigned long) = 0;
  virtual int setopt(int, ForwarderIF*) = 0;
  int transmit_packet(char *, int n);

protected:
  static VR* vr;
};


/*
 * UDP Socket
 */ 

class UDPSock : public Sock, public ProtocolStack
{
public:
  UDPSock(VR*);
  UDPSock(int m);
  ~UDPSock();
  int process(PSPacket*);
  int dispatch(Packet *p);
  int read(char*, int);
  int write(char*, int);
  int setopt(int, ForwarderIF*);
  int setopt(int a, int b){return setopt(a,(unsigned long)b);}
  int setopt(int, unsigned long);
  int get_fh(){return sv[1];}

private:
  int localport, remoteport;
  static int number_of_sockets;
  static UDPSock *socks[MAX_NUMBER_OF_UDP_SOCKETS];
  int socks_index;
  int mode;
  int sv[2];
  unsigned long d_addr;
  unsigned long s_addr;
  int ttl;
  int bytes_in_socket;
  int hdr_include;
  ForwarderIF *forwarder;
  int blocking;
};

/*
 * Raw Socket
 */ 

class RawSock : public Sock
{
public:
  RawSock(VR*);			// called by the forwarder
  RawSock(int mode); // proto: icmp, mode:1 read, mode:2 write
  ~RawSock();

  int dispatch(Packet *p);
  int read(char*, int);
  int write(char*, int);
  int setopt(int, ForwarderIF*);
  int setopt(int a, int b){return setopt(a,(unsigned long)b);}
  int setopt(int, unsigned long);
  int get_fh(){return sv[1];}

private:
  static int number_of_sockets;
  static RawSock *socks[MAX_NUMBER_OF_RAW_SOCKETS];
  unsigned int proto;
  int socks_index;
  int sv[2];
  int mode;
  unsigned long d_addr;
  unsigned long s_addr;
  int ttl;
  int bytes_in_socket;
  int hdr_include;
  ForwarderIF *forwarder;
  int blocking;
};



#endif

