#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "vr.H"
#include "tools.H"
#include "interface.H"
#include "ioc.H"
#include "udp.H"
#include "icmp.H"
#include "tunnel_endpoint.H"
#include "routing.H"
#include "shell.H"
#include "tools.H"
class Forwarder;


VR::VR()
{
  num_of_interfaces=0;
  highest_interface_idx=0;
  startup_time=gettimeasdouble();
  cfh=1;

  // default values


  ttl_default=64;		// ttl default value

  // shell related things 

  shell_binary=0;		// shell is human readable
  shell_port=0;			// shell is on console
  shell_password=NULL;		// no passwort check
  shell_password_tries=3;	// number of passwort tries
  shell_startupscript=NULL;	// no shell startup script

  // paths and filenames

  lob_path="../lib";		// path to the precompiled objects
  lob_extension=".vr";		// extension of th lob files (.vr)
  fifo_basename="/var/tmp/varlink";
  softlink_basename="/dev/sol";

  // queueing components

  if_bandwidth=1000000/8;	// interface bandwidth in Mbps
  if_bucketsize=2048;		// interface bucketsize 2K
  tbf_bandwidth=1000000/8;	// tbf default bandwidth in Mbps
  tbf_bucketsize=2048;		// tbf default bucketsize 2K
  dpt_queuelen=64;		// 64 packets in the queue 

  // ip translation

  iptrans_enable=0;		// ip translation disabled
  iptrans_rx_address=iptrans_rx_netmask=char2n("255.255.255.255");
  iptrans_rx_value=iptrans_rx_mask=0;	
  iptrans_tx_address=iptrans_tx_netmask=char2n("255.255.255.255");
  iptrans_tx_value=iptrans_tx_mask=0;
}


int VR::init()
{
  vrsched   = new VRsched(this);
  udpstack  = new UDPstack(this);
  icmpstack = new ICMP(this);
  tunnel_ep = new Tunnel_Endpoint(this);
  pf        = new Forwarder(this);
  routing   = new Routing(this);
  shell     = new Shell(this);
  return 0;
};

double VR::uptime()
{
  return gettimeasdouble()-startup_time;
}

int VR::run()
{
  vrsched->core();
  return 0;
}

VR::~VR()
{
  delete shell;
  delete routing;
  delete pf;
  delete vrsched;
  delete udpstack;
  delete icmpstack;
  return;
}


Interface* VR::get_if(char *n){
  for(int i=0; i<=highest_interface_idx ;++i){
    if(!strcmp(in[i]->id(),n))return in[i];
  }
  return NULL;
}


int VR::add_interface(Interface *ifa)
{
  int i;

  for(i=0;i<MAX_NOF_INTERFACES;++i)
    if(in[i]==NULL){
      in[i]=ifa;
      num_of_interfaces++;
      break;
    }
  
  if(i>highest_interface_idx)highest_interface_idx=i;
  return i;
}

int VR::delete_interface(int ifc)
{
  if(in[ifc]==NULL)return -1;
  delete in[ifc];
  in[ifc]=NULL;
  num_of_interfaces--;
  if(ifc==highest_interface_idx)
    highest_interface_idx=ifc-1;

  // clean up routes
  routing->cleanup();
  
  return 0;
}



int VR::configurebyfile(char *file)
{
  FILE *fp=fopen(file,"r");
 
 if(!fp){
    fprintf(stderr,"cannot read configfile %s\n",file);
    exit(0);
  }
  
  char buffer[1024];

  while(fgets(buffer,1024,fp)){
    strip_trailing(buffer,"\n\r\t ");
    char *c=configlinesplit(buffer);
    if(!c)continue;

    if(!strcmp(buffer,"include")){
      configurebyfile(c);
      continue;
    }


    if(!strcmp(buffer,"binary_shell")){
      shell_binary=istrue(c);
      continue;
    }
    if(!strcmp(buffer,"port_of_shell")){
      shell_port=atoi(c);
      continue;
    }
    if(!strcmp(buffer,"hostname_for_shell")){
      shell_hostname=strdup(c);
      continue;
    }
    if(!strcmp(buffer,"lob_path")){
      lob_path=strdup(c);
      continue;
    }
    if(!strcmp(buffer,"fifo_basename")){
      fifo_basename=strdup(c);
      continue;
    }
    if(!strcmp(buffer,"softlink_basename")){
      softlink_basename=strdup(c);
      continue;
    }
    if(!strcmp(buffer,"shell_startup")){
      shell_startupscript=strdup(c);
      continue;
    }
    if(!strcmp(buffer,"if_bandwidth")){
      if_bandwidth=(unsigned long)(atof(c)*1000000.0/8.0);
      continue;
    }

    if(!strcmp(buffer,"if_bucketsize")){
      if_bucketsize=atol(c);
      continue;
    }

    if(!strcmp(buffer,"tbf_bandwidth")){
      tbf_bandwidth=(unsigned long)(atof(c)*1000000.0/8.0);
      continue;
    }

    if(!strcmp(buffer,"tbf_bucketsize")){
      tbf_bucketsize=atol(c);
      continue;
    }

    if(!strcmp(buffer,"dpt_queuelen")){
      dpt_queuelen=atoi(c);
      continue;
    }

    if(!strcmp(buffer,"shell_password")){
      shell_password=strdup(c);
      continue;
    }

    if(!strcmp(buffer,"shell_password_tries")){
      shell_password_tries=atoi(c);
      continue;
    }

    if(!strcmp(buffer,"iptrans_enable")){
      iptrans_enable=istrue(c);
      continue;
    }

    if(!strcmp(buffer,"iptrans_rx_address")){
      iptrans_rx_address=char2n(c);
      continue;
    }

    if(!strcmp(buffer,"iptrans_tx_address")){
      iptrans_tx_address=char2n(c);
      continue;
    }

    if(!strcmp(buffer,"iptrans_rx_netmask")){
      iptrans_rx_netmask=char2n(c);
      continue;
    }

    if(!strcmp(buffer,"iptrans_tx_netmask")){
      iptrans_tx_netmask=char2n(c);
      continue;
    }

    if(!strcmp(buffer,"iptrans_rx_value")){
      iptrans_rx_value=char2n(c);
      continue;
    }

    if(!strcmp(buffer,"iptrans_tx_value")){
      iptrans_tx_value=char2n(c);
      continue;
    }

    if(!strcmp(buffer,"iptrans_rx_mask")){
      iptrans_rx_mask=char2n(c);
      continue;
    }

    if(!strcmp(buffer,"iptrans_tx_mask")){
      iptrans_tx_mask=char2n(c);
      continue;
    }

    if(!strcmp(buffer,"ttl_default")){
      ttl_default=atoi(c);
      continue;
    }

    fhprintf(cfh,"unkown keyword '%s'\n",buffer);
  }
  fclose(fp);

#ifdef DEBUG_CONFIG
  fhprintf(cfh,"DEBUG_CONFIG: iptrans_enable: %d\n",iptrans_enable);
  fhprintf(cfh,"DEBUG_CONFIG: pausing 5 seconds\n");
  sleep(5);
#endif

  return 0;
}

int VR::set_startupfile(char *f)
{
  shell_startupscript=strdup(f);
  return 0;
}


int VR::set_portnumber(int i)
{
  shell_port=i;
  return 0;
}


int VR::set_hostname(char *s)
{
  shell_hostname=strdup(s);
  return 0;
}

int VR::set_fifobase(char *s)
{
  fifo_basename=strdup(s);
  return 0;
}

int VR::set_lob_path(char *s)
{
  lob_path=strdup(s);
  return 0;
}
